/*
 * Copyright (c) 2001, Thierry Van Elsuw 
 * All rights reserved. 
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions 
 * are met:
 * 
 * Redistributions of source code must retain the above copyright 
 * notice, this list of conditions and the following disclaimer.
 * 
 * Redistributions in binary form must reproduce the above copyright 
 * notice, this list of conditions and the following disclaimer in the 
 * documentation and/or other materials provided with the distribution.
 * 
 * Neither the name of DS Improve nor the names of its contributors may 
 * be used to endorse or promote products derived from this software 
 * without specific prior written permission.
 * 
 * Module name   : tveClock.cpp
 * Creation date : 14/02/2001
 * e-mail        : Thierry Van Elsuwe <tvelsuwe@swing.be>
 */

#include "voyager.h"
#include "tveClock.h"


/*
 *
 */
tveClock::tveClock()
{
    start_tick    = SDL_GetTicks();
    last_tick     = start_tick;
    last_frame    = 0;
    current_frame = 0;
    current_fps   = 0;
}


/*
 *
 */
void tveClock::startClock()
{
    start_tick = SDL_GetTicks();
}


/*
 *
 */
Uint32 tveClock::getClock()
{
    return SDL_GetTicks()-start_tick;
}


/*
 *
 */
Uint32 tveClock::getCurrentFPS()
{
    Uint32 current_tick = SDL_GetTicks();

    // refresh if the interval is above 2 sec
    if (current_tick-last_tick > 2000)
    {
        current_fps = (1000*(current_frame-last_frame))/(current_tick-last_tick);
        last_tick = current_tick;
        last_frame = current_frame;
    }
    return current_fps;
}


/*
 *
 */
Uint32 tveClock::getTotalFPS()
{
    return (1000*current_frame)/(SDL_GetTicks()-start_tick);
}


/*
 *
 */
void tveClock::incFrame()
{
    current_frame++;
}


/*
 *
 */
void tveClock::resetFrame()
{
    last_frame    = 0;
    current_frame = 0;
}
