/*
 * Copyright (c) 2001, Thierry Van Elsuw 
 * All rights reserved. 
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions 
 * are met:
 * 
 * Redistributions of source code must retain the above copyright 
 * notice, this list of conditions and the following disclaimer.
 * 
 * Redistributions in binary form must reproduce the above copyright 
 * notice, this list of conditions and the following disclaimer in the 
 * documentation and/or other materials provided with the distribution.
 * 
 * Neither the name of DS Improve nor the names of its contributors may 
 * be used to endorse or promote products derived from this software 
 * without specific prior written permission.
 * 
 * Module name   : tveDemo.cpp
 * Creation date : 09/03/2001
 * e-mail        : Thierry Van Elsuwe <tvelsuwe@swing.be>
 */

#include "voyager.h"
#include "tveDemo.h"


/*
 *
 */
tveDemo::tveDemo()
{
    state = DEMO_STOP;
}


/*
 *
 */
void tveDemo::start(int state, char *demoname)
{
    if (this->state!=DEMO_STOP)
    {
        delete []pos;
        free(this->demoname);
    }
    this->state = state;
    this->demoname = strdup(demoname);
    if (state==DEMO_RECORDING)
    {
        pos = new tvePosition[MAX_DEMO_SIZE];
        pos[0].pos.x  = tveVoyager::cv_posx->value.floating;
        pos[0].pos.y  = tveVoyager::cv_posy->value.floating;
        pos[0].pos.z  = tveVoyager::cv_posz->value.floating;
        pos[0].anglex = tveVoyager::cv_anglex->value.floating;
        pos[0].angley = tveVoyager::cv_angley->value.floating;
        pos[0].tick   = my_clock->getClock();
    }
    else if ( (state==DEMO_PLAYING) || (state==DEMO_LOOPING) )
    {
        gzFile f = tveVoyager::openFile(demoname, "rb");

        if (!f)
        {
            console->printf("cannot start demo %s (file not found)", demoname);
            this->state = DEMO_STOP;
            return;
        }
        pos = new tvePosition[MAX_DEMO_SIZE];
        gzread(f, pos, sizeof(tvePosition)*MAX_DEMO_SIZE);
        gzclose(f);
        loop = (state==DEMO_LOOPING);
        this->state = DEMO_PLAYING;
    }
    cur_pos = 0;
}


/*
 *
 */
void tveDemo::execute()
{
    if (state==DEMO_PLAYING)
    {
        if (pos[cur_pos].pos.y > -10000.0f) // for now, play as fast as possible
        {
            tveVoyager::cv_posx->value.floating = pos[cur_pos].pos.x;
            tveVoyager::cv_posy->value.floating = pos[cur_pos].pos.y;
            tveVoyager::cv_posz->value.floating = pos[cur_pos].pos.z;
            tveVoyager::cv_anglex->value.floating = pos[cur_pos].anglex;
            tveVoyager::cv_angley->value.floating = pos[cur_pos].angley;
            cur_pos++;
        }
        else if (loop)
        {
            cur_pos = 0;
        }
    }
    else if (state==DEMO_RECORDING)
    {
        // check if we have moved since last frame
        if ( (tveVoyager::cv_posx->value.floating!=pos[cur_pos].pos.x) ||
             (tveVoyager::cv_posy->value.floating!=pos[cur_pos].pos.y) ||
             (tveVoyager::cv_posz->value.floating!=pos[cur_pos].pos.z) ||
             (tveVoyager::cv_anglex->value.floating!=pos[cur_pos].anglex) ||
             (tveVoyager::cv_angley->value.floating!=pos[cur_pos].angley) )
        {
            cur_pos++;
            pos[cur_pos].pos.x  = tveVoyager::cv_posx->value.floating;
            pos[cur_pos].pos.y  = tveVoyager::cv_posy->value.floating;
            pos[cur_pos].pos.z  = tveVoyager::cv_posz->value.floating;
            pos[cur_pos].anglex = tveVoyager::cv_anglex->value.floating;
            pos[cur_pos].angley = tveVoyager::cv_angley->value.floating;
            pos[cur_pos].tick   = my_clock->getClock();
        }
    }
}


/*
 *
 */
void tveDemo::togglePause()
{
    if (state == DEMO_PAUSE)
    {
        state = DEMO_PLAYING;
    }
    else if (state != DEMO_STOP)
    {
        state = DEMO_PAUSE;
    }
}


/*
 *
 */
void tveDemo::stop()
{
    if (state == DEMO_RECORDING)
    {
        gzFile f = tveVoyager::openFile(demoname, "wb");

        if (!f)
        {
            console->printf("cannot save demo %s", demoname);
        }
        else
        {
            cur_pos++;
            pos[cur_pos].pos.y  = -10001.0f; // EOD marker
            for (int i=0; i<=cur_pos; i++)
            {
                gzwrite(f, &pos[i], sizeof(tvePosition));
            }
            gzclose(f);
        }
    }
    state = DEMO_STOP;
    delete []pos;
    free(demoname);
}
