/*
 * Copyright (c) 2001, Thierry Van Elsuw 
 * All rights reserved. 
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions 
 * are met:
 * 
 * Redistributions of source code must retain the above copyright 
 * notice, this list of conditions and the following disclaimer.
 * 
 * Redistributions in binary form must reproduce the above copyright 
 * notice, this list of conditions and the following disclaimer in the 
 * documentation and/or other materials provided with the distribution.
 * 
 * Neither the name of DS Improve nor the names of its contributors may 
 * be used to endorse or promote products derived from this software 
 * without specific prior written permission.
 * 
 * Module name   : tveHeightmap.h
 * Creation date : 23/02/2001
 * e-mail        : Thierry Van Elsuwe <tvelsuwe@swing.be>
 */

#ifndef _TVEHEIGHTMAP_H_
#define _TVEHEIGHTMAP_H_

#include "voyager.h"
#include "tveTexture.h"


enum tveTexName
{
    SHADOW = 0,
    WATER,
    GRASS = WATER+16, // reserve 16 textures for water animation
    ROCK,
    SNOW,
    DIRT,
    TREE,
    SKY,
    NUM_TEX
};


enum tveGeoMap  // not yet used 
{
    GEO_TREE    = 0x01,
    GEO_FIR     = 0x02,
    GEO_FLOWER  = 0x04,
    GEO_RIVER   = 0x08,
    GEO_ROAD    = 0x10,
    GEO_ROCK    = 0x20,
    GEO_SPECIAL = 0x40,
};


class tveHeightMap
{
private:
    void loadFromTGN(gzFile f);
    void loadFromH16(gzFile f);
    void loadFromTGA(gzFile f);
    void saveTGA();
    void recurse0(int left, int top, int right, int bottom);
    void recurse1(int left, int top, int right, int bottom);
    void recurse2(int left, int top, int right, int bottom);
    GLuint generateShadowTex();
    void generateGeoMap();
    void generateCoarseLevel();
public:
    Uint16 height;
    Uint16 width;
    Sint16 detail[MAXX+1][MAXZ+1];
    Sint16 coarse[MAXX/2+1][MAXZ/2+1];
    Uint8  geomap[MAXX+1][MAXZ+1];
    GLuint *tex;
    tveHeightMap();
    ~tveHeightMap();
    void loadFromFile(char *filename);
    void generate(int height, int width, int base_y, int method);
    void smooth();
};

#endif
