/*
 * Copyright (c) 2001, Thierry Van Elsuw 
 * All rights reserved. 
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions 
 * are met:
 * 
 * Redistributions of source code must retain the above copyright 
 * notice, this list of conditions and the following disclaimer.
 * 
 * Redistributions in binary form must reproduce the above copyright 
 * notice, this list of conditions and the following disclaimer in the 
 * documentation and/or other materials provided with the distribution.
 * 
 * Neither the name of DS Improve nor the names of its contributors may 
 * be used to endorse or promote products derived from this software 
 * without specific prior written permission.
 * 
 * Module name   : tveLandscape.cpp
 * Creation date : 16/02/2001
 * e-mail        : Thierry Van Elsuwe <tvelsuwe@swing.be>
 */

#include "voyager.h"
#include "tveLandscape.h"
#include "tveTexture.h"

#ifdef DEMO_DS
#define NUM_MAPS 8
char *mapname[NUM_MAPS] = {
    "Little River.tve",
    "River One.tve",
    "River Two.tve",
    "The Island.tve",
    "tux.tve", 
    "terrain0.tve", 
    "terrain1.tve", 
    "terrain2.tve", 
};
#else
#define NUM_MAPS 14
char *mapname[NUM_MAPS] = {
    "1st Nice Mountain.tve",
    "Four Big Mountains.tve",
    "Land of Shadows.tve",
    "Little River.tve",
    "River One.tve",
    "River Two.tve",
    "Rocks and Water.tve",
    "Rough Hill.tve",
    "The Big One.tve",
    "The Three Isles.tve",
    "The Island.tve",
    "Two Big Mountains.tve",
    "Wide Area.tve",
    "terrain.ter", 
};
#endif


/*
 *
 */
tveLandscape::tveLandscape(char *filename)
{
    console->printf("loading heightmap...");
    height_map = new tveHeightMap();
    if (filename)
    {
        int num = atoi(filename);

        if (num>0)
        {
            if (num<=NUM_MAPS) 
            {
                height_map->loadFromFile(mapname[num-1]);
            }
            else
            {
                height_map->loadFromFile(mapname[RANDOM(NUM_MAPS)]);
            }
        }
        else
        {
            height_map->loadFromFile(filename);
        }
    }
    else
    {
        height_map->loadFromFile(mapname[RANDOM(NUM_MAPS)]);
    }
    console->printf("Generating Quad-Tree...");
    quad_tree = new tveQuadTree(height_map);
    console->printf("Generating Trees...");
    generateTreeList();
    console->printf("DONE.\n");
}


/*
 *
 */
tveLandscape::~tveLandscape()
{
    glDeleteLists(list_tree, 1);
    delete height_map;
    delete quad_tree;
}


/*
 *
 */
void tveLandscape::generateTreeList()
{
    list_tree = glGenLists(1);
    glNewList(list_tree, GL_COMPILE);

    glActiveTextureARB(GL_TEXTURE1_ARB);
    glDisable(GL_TEXTURE_2D);

    glActiveTextureARB(GL_TEXTURE0_ARB);
    glEnable(GL_TEXTURE_2D);
    glBindTexture(GL_TEXTURE_2D, height_map->tex[TREE]);

    glColor4f(1.0f, 1.0f, 1.0f, 1.0f);
    glBegin(GL_TRIANGLE_STRIP);
      glMultiTexCoord2fARB(GL_TEXTURE0_ARB, 0.0f, 0.0f);
      glVertex3f(0.0f, 0.0f, 1.0f);
      glMultiTexCoord2fARB(GL_TEXTURE0_ARB, 0.0f, 1.0f);
      glVertex3f(0.0f, 3.0f, 1.0f);
      glMultiTexCoord2fARB(GL_TEXTURE0_ARB, 1.0f, 0.0f);
      glVertex3f(2.0f, 0.0f, 1.0f);
      glMultiTexCoord2fARB(GL_TEXTURE0_ARB, 1.0f, 1.0f);
      glVertex3f(2.0f, 3.0f, 1.0f);
    glEnd();
    glBegin(GL_TRIANGLE_STRIP);
      glMultiTexCoord2fARB(GL_TEXTURE0_ARB, 0.0f, 0.0f);
      glVertex3f(1.0f, 0.0f, 0.0f);
      glMultiTexCoord2fARB(GL_TEXTURE0_ARB, 0.0f, 1.0f);
      glVertex3f(1.0f, 3.0f, 0.0f);
      glMultiTexCoord2fARB(GL_TEXTURE0_ARB, 1.0f, 0.0f);
      glVertex3f(1.0f, 0.0f, 2.0f);
      glMultiTexCoord2fARB(GL_TEXTURE0_ARB, 1.0f, 1.0f);
      glVertex3f(1.0f, 3.0f, 2.0f);
    glEnd();

    glEndList();
}


/*
 *
 */
void tveLandscape::drawTrees()
{
    quad_tree->renderGeoMap(true);
}


/*
 *
 */
void tveLandscape::drawWater()
{
    quad_tree->renderWater();
}


/*
 *
 */
void tveLandscape::drawSky()
{
    glDisable(GL_BLEND);
    glBindTexture(GL_TEXTURE_2D, height_map->tex[SKY]);
    glColor3f(1.0f, 1.0f, 1.0f);
    glBegin(GL_TRIANGLE_FAN);
      float wind = quad_tree->wind;
      glTexCoord2f(1.0f+wind, 1.0f+wind);  glVertex3f( 0.0f, height_map->height/4.0f, 0.0f);
      glTexCoord2f(2.0f+wind, 2.0f+wind);  glVertex3f( height_map->width*2.0f, 0.0f, height_map->height*2.0f);
      glTexCoord2f(2.0f+wind, 0.0f+wind);  glVertex3f( height_map->width*2.0f, 0.0f,-height_map->height*2.0f);
      glTexCoord2f(0.0f+wind, 0.0f+wind);  glVertex3f(-height_map->width*2.0f, 0.0f,-height_map->height*2.0f);
      glTexCoord2f(0.0f+wind, 2.0f+wind);  glVertex3f(-height_map->width*2.0f, 0.0f, height_map->height*2.0f);
      glTexCoord2f(2.0f+wind, 2.0f+wind);  glVertex3f( height_map->width*2.0f, 0.0f, height_map->height*2.0f);
    glEnd();
    glEnable(GL_BLEND);
}


/*
 *
 */
void tveLandscape::drawQuadTreeVertexArray(bool lock)
{
    quad_tree->wind += 0.000004f * tveVoyager::ratio;
    if ( quad_tree->wind > 1.0f )
    {
        quad_tree->wind -= 1.0f;
    }
    quad_tree->renderTerrain(lock);
}


/*
 *
 */
GLfloat tveLandscape::getYPos(GLfloat x, GLfloat z)
{
    GLfloat res = tveVoyager::cv_water->value.floating/SCALE_Y;
    int xpos = (int)floor(x*0.5f);
    int zpos = (int)floor(z*0.5f);

    if ((xpos < height_map->width) && (xpos > 0) && (zpos < height_map->height) && (zpos > 0))
    {
        res = MAX(res, height_map->detail[xpos  ][zpos  ]);
        res = MAX(res, height_map->detail[xpos+1][zpos  ]);
        res = MAX(res, height_map->detail[xpos  ][zpos+1]);
        res = MAX(res, height_map->detail[xpos+1][zpos+1]);
    }
    return res*SCALE_Y;
}
