/*
 * Copyright (c) 2001, Thierry Van Elsuw 
 * All rights reserved. 
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions 
 * are met:
 * 
 * Redistributions of source code must retain the above copyright 
 * notice, this list of conditions and the following disclaimer.
 * 
 * Redistributions in binary form must reproduce the above copyright 
 * notice, this list of conditions and the following disclaimer in the 
 * documentation and/or other materials provided with the distribution.
 * 
 * Neither the name of DS Improve nor the names of its contributors may 
 * be used to endorse or promote products derived from this software 
 * without specific prior written permission.
 * 
 * Module name   : tveModel.h
 * Creation date : 21/04/2001
 * e-mail        : Thierry Van Elsuwe <tvelsuwe@swing.be>
 */

#ifndef _TVEMODEL_H_
#define _TVEMODEL_H_


class tveGenericModel
{
protected:
    GLuint tex;
public:
    tveGenericModel();
    virtual ~tveGenericModel();
    virtual void print() = 0;
    virtual void render() = 0;
};


//===================================================================


typedef struct 
{ 
   Sint32 magic;
   Sint32 version;
   Sint32 skin_width;
   Sint32 skin_height;
   Sint32 frame_size;
   Sint32 num_skins;
   Sint32 num_vertices;
   Sint32 num_tex_coords;
   Sint32 num_polygons;
   Sint32 num_gl_commands;
   Sint32 num_frames;
   Sint32 offset_skins;
   Sint32 offset_tex_coords;
   Sint32 offset_polygons;
   Sint32 offset_frames;
   Sint32 offset_gl_commands;
   Sint32 offset_end;
} tveMdzHeader;


typedef char tveMdzSkin[32];


typedef struct
{
   Sint16 s;
   Sint16 t;
} tveMdzTexCoord;


typedef struct
{
   Sint16 vertex_indices[3];
   Sint16 texture_indices[3];
} tveMdzPolygons;


typedef struct
{
   float vertex[3];
   float normal[3];
} tveMdzPolygonsVertex;


typedef struct
{
   char name[16];
   tveMdzPolygonsVertex *vertices;
} tveMdzFrame;


class tveMdzModel : public tveGenericModel
{
private:
    tveMdzHeader    header;
    tveMdzSkin      *skins;
    tveMdzTexCoord  *tex_coords;
    tveMdzPolygons  *polygons;
    tveMdzFrame     *frames;
    Sint32          *gl_command_buffer;
    void            loadSkin(gzFile f);
    float           frame_num;
public:
    void print();
    void render();
    tveMdzModel(gzFile f);
    virtual ~tveMdzModel();
};


//===================================================================


typedef struct
{
    int material_index;
    int num_triangles;
    int *triangle_indices;
} tveMS3DMesh;


typedef struct
{
    float   ambient[4];
    float   diffuse[4]; 
    float   specular[4];
    float   emissive[4];
    float   shininess;
    GLuint  tex;
    char    *texture_filename;
} tveMS3DMaterial;


typedef struct
{
    float vertex_normals[3][3];
    float s[3];
    float t[3];
    int   vertex_indices[3];
} tveMS3DTriangle;


typedef struct
{
    char    bone_id;    // for skeletal animation
    float   location[3];
} tveMS3DVertex;


class tveMS3DModel : public tveGenericModel
{
private:
	int             num_meshes;
	tveMS3DMesh     *meshes;
	int             num_materials;
	tveMS3DMaterial *materials;
	int             num_triangles;
	tveMS3DTriangle *triangles;
	int             num_vertices;
	tveMS3DVertex   *vertices;
public:
    tveMS3DModel(gzFile f);
    virtual ~tveMS3DModel();
    void print();
    void render();
	bool loadModelData(char *filename);
	void draw();
	void reloadTextures();
};

//===================================================================


class tveModel
{
private:
    tveGenericModel *model;
    tveVertex       *getBoundingBox();  // not yet used
public:
    tveModel(char *filename);
    ~tveModel();
    void print();
    void render();
};


#endif
