/*
 * Copyright (c) 2001, Thierry Van Elsuw 
 * All rights reserved. 
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions 
 * are met:
 * 
 * Redistributions of source code must retain the above copyright 
 * notice, this list of conditions and the following disclaimer.
 * 
 * Redistributions in binary form must reproduce the above copyright 
 * notice, this list of conditions and the following disclaimer in the 
 * documentation and/or other materials provided with the distribution.
 * 
 * Neither the name of DS Improve nor the names of its contributors may 
 * be used to endorse or promote products derived from this software 
 * without specific prior written permission.
 * 
 * Module name   : tvePattern.cpp
 * Creation date : 16/02/2001
 * e-mail        : Thierry Van Elsuwe <tvelsuwe@swing.be>
 */

#include "voyager.h"
#include "tvePattern.h"
#include "tveQuadtree.h"


//=====================================================================
// tvePattern
//=====================================================================

/*
 *
 */
tvePattern::tvePattern()
{
}


//=====================================================================
// tveVertexArray
//=====================================================================

/*
 *
 */
tveVertexArray::tveVertexArray()
{
}


/*
 *
 */
tveVertexArray::~tveVertexArray()
{
}


/*
 *
 */
void tveVertexArray::drawArray(int dir)
{
    if (detail) // full detail
    {
        for (int k=0; k<LEAF_GRANULARITY; k++)
        {
            glDrawElements(GL_TRIANGLE_STRIP, (LEAF_GRANULARITY+1)*2, GL_UNSIGNED_INT, &faces_strip_detail[k*(LEAF_GRANULARITY+1)*2]);
        }
    }
    else if (!dir) // full coarse
    {
        for (int k=0; k<LEAF_GRANULARITY; k+=2)
        {
            glDrawElements(GL_TRIANGLE_STRIP, (LEAF_GRANULARITY/2+1)*2, GL_UNSIGNED_INT, &faces_strip_coarse[(k/2)*(LEAF_GRANULARITY/2+1)*2]);
        }
    }
    else // mix of coarse and detail
    {
        int offset_x = 0;
        int offset_y = 0;
        int size_x   = LEAF_GRANULARITY/2+1;
        int size_y   = LEAF_GRANULARITY/2+1;

        if (dir & DIR_NORTH_MASK)
        {
            size_y--;
            offset_y++;
        }
        if (dir & DIR_SOUTH_MASK)
        {
            size_y--;
        }
        if (dir & DIR_WEST_MASK)
        {
            size_x--;
            offset_x++;
        }
        if (dir & DIR_EAST_MASK)
        {
            size_x--;
        }
        for (int k=offset_y; k<size_y+offset_y-1; k++)
        {
            glDrawElements(GL_TRIANGLE_STRIP, size_x*2, GL_UNSIGNED_INT, &faces_strip_coarse[2*k*(LEAF_GRANULARITY/2+1)+offset_x*2]);
        }
        // render junctions
        for (int j=2; j<LEAF_GRANULARITY-2; j+=2)
        {
            if (dir & DIR_NORTH_MASK)
            {
                glBegin(GL_TRIANGLE_STRIP);
                  glArrayElement(j+0);
                  glArrayElement(2*(LEAF_GRANULARITY+1)+0+j);
                  glArrayElement(j+1);
                  glArrayElement(2*(LEAF_GRANULARITY+1)+2+j);
                  glArrayElement(j+2);
                glEnd();
            }
            if (dir & DIR_WEST_MASK)
            {
                glBegin(GL_TRIANGLE_STRIP);
                  glArrayElement((LEAF_GRANULARITY+1)*(j+2));
                  glArrayElement((LEAF_GRANULARITY+1)*(j+2)+2);
                  glArrayElement((LEAF_GRANULARITY+1)*(j+1));
                  glArrayElement((LEAF_GRANULARITY+1)*j+2);
                  glArrayElement((LEAF_GRANULARITY+1)*j);
                glEnd();
            }
            if (dir & DIR_EAST_MASK)
            {
                glBegin(GL_TRIANGLE_STRIP);
                  glArrayElement(LEAF_GRANULARITY-2+(LEAF_GRANULARITY+1)*j+2);
                  glArrayElement(LEAF_GRANULARITY-2+(LEAF_GRANULARITY+1)*j);
                  glArrayElement(LEAF_GRANULARITY-2+(LEAF_GRANULARITY+1)*(j+1)+2);
                  glArrayElement(LEAF_GRANULARITY-2+(LEAF_GRANULARITY+1)*(j+2));
                  glArrayElement(LEAF_GRANULARITY-2+(LEAF_GRANULARITY+1)*(j+2)+2);
                glEnd();
            }
            if (dir & DIR_SOUTH_MASK)
            {
                glBegin(GL_TRIANGLE_STRIP);
                  glArrayElement((LEAF_GRANULARITY+1)*(LEAF_GRANULARITY+1)-(j+0)-1);
                  glArrayElement((LEAF_GRANULARITY+1)*(LEAF_GRANULARITY+1)-(2*(LEAF_GRANULARITY+1)+0+j)-1);
                  glArrayElement((LEAF_GRANULARITY+1)*(LEAF_GRANULARITY+1)-(j+1)-1);
                  glArrayElement((LEAF_GRANULARITY+1)*(LEAF_GRANULARITY+1)-(2*(LEAF_GRANULARITY+1)+2+j)-1);
                  glArrayElement((LEAF_GRANULARITY+1)*(LEAF_GRANULARITY+1)-(j+2)-1);
                glEnd();
            }
        }
        if (dir & (DIR_NORTH_MASK|DIR_WEST_MASK))
        {
            glBegin(GL_TRIANGLE_FAN);
              glArrayElement(2*(LEAF_GRANULARITY+1)+2);
              glArrayElement(2);
              if (dir & DIR_NORTH_MASK) glArrayElement(1);
              glArrayElement(0);
              if (dir & DIR_WEST_MASK) glArrayElement(LEAF_GRANULARITY+1+0);
              glArrayElement(2*(LEAF_GRANULARITY+1)+0);
            glEnd();
        }
        if (dir & (DIR_EAST_MASK|DIR_NORTH_MASK))
        {
            glBegin(GL_TRIANGLE_FAN);
              glArrayElement(LEAF_GRANULARITY-2+2*(LEAF_GRANULARITY+1)+0);
              glArrayElement(LEAF_GRANULARITY-2+2*(LEAF_GRANULARITY+1)+2);
              if (dir & DIR_EAST_MASK) glArrayElement(LEAF_GRANULARITY-2+LEAF_GRANULARITY+1+2);
              glArrayElement(LEAF_GRANULARITY-2+2);
              if (dir & DIR_NORTH_MASK) glArrayElement(LEAF_GRANULARITY-2+1);
              glArrayElement(LEAF_GRANULARITY-2+0);
            glEnd();
        }
        if (dir & (DIR_WEST_MASK|DIR_SOUTH_MASK))
        {
            glBegin(GL_TRIANGLE_FAN);
              glArrayElement((LEAF_GRANULARITY+1)*(LEAF_GRANULARITY-2)+2);
              glArrayElement((LEAF_GRANULARITY+1)*(LEAF_GRANULARITY-2)+0);
              if (dir & DIR_WEST_MASK) glArrayElement((LEAF_GRANULARITY+1)*(LEAF_GRANULARITY-1)+0);
              glArrayElement((LEAF_GRANULARITY+1)*(LEAF_GRANULARITY+0)+0);
              if (dir & DIR_SOUTH_MASK) glArrayElement((LEAF_GRANULARITY+1)*(LEAF_GRANULARITY+0)+1);
              glArrayElement((LEAF_GRANULARITY+1)*(LEAF_GRANULARITY+0)+2);
            glEnd();
        }
        if (dir & (DIR_SOUTH_MASK|DIR_EAST_MASK))
        {
            glBegin(GL_TRIANGLE_FAN);
              glArrayElement((LEAF_GRANULARITY+1)*(LEAF_GRANULARITY-1)-3);
              glArrayElement((LEAF_GRANULARITY+1)*(LEAF_GRANULARITY+1)-3);
              if (dir & DIR_SOUTH_MASK) glArrayElement((LEAF_GRANULARITY+1)*(LEAF_GRANULARITY+1)-2);
              glArrayElement((LEAF_GRANULARITY+1)*(LEAF_GRANULARITY+1)-1);
              if (dir & DIR_EAST_MASK) glArrayElement((LEAF_GRANULARITY+1)*(LEAF_GRANULARITY+0)-1);
              glArrayElement((LEAF_GRANULARITY+1)*(LEAF_GRANULARITY-1)-1);
            glEnd();
        }
    }
}


/*
 *
 */
void tveVertexArray::render(float y, int dir)
{
    glEnableClientState(GL_VERTEX_ARRAY);
    glEnableClientState(GL_TEXTURE_COORD_ARRAY);

    //first pass
    glClientActiveTextureARB(GL_TEXTURE0_ARB);
    glTexCoordPointer(2, GL_FLOAT, 0, tex_coord[0]);
    glActiveTextureARB(GL_TEXTURE0_ARB);
    glEnable(GL_TEXTURE_2D);
    glBindTexture(GL_TEXTURE_2D, height_map->tex[SHADOW]);

    glClientActiveTextureARB(GL_TEXTURE1_ARB);
    glTexCoordPointer(2, GL_FLOAT, 0, tex_coord[1]);
    glActiveTextureARB(GL_TEXTURE1_ARB);
    glEnable(GL_TEXTURE_2D);
    glBindTexture(GL_TEXTURE_2D, height_map->tex[GRASS]);

    glVertexPointer(3, GL_FLOAT, 0, vertices);
    //TODO
    if (top->lock) glLockArraysEXT(0, (LEAF_GRANULARITY+1)*(LEAF_GRANULARITY+1));
    drawArray(dir);

    // prepare second and third pass
    glEnableClientState(GL_COLOR_ARRAY);
    glClientActiveTextureARB(GL_TEXTURE0_ARB);
    glTexCoordPointer(2, GL_FLOAT, 0, tex_coord[1]);
    glActiveTextureARB(GL_TEXTURE1_ARB);
    glDisable(GL_TEXTURE_2D);
    glActiveTextureARB(GL_TEXTURE0_ARB);
    glDepthMask(0);

    //second pass
    //TODO
    if (y > 128.0f)
    {
        glColorPointer(4, GL_UNSIGNED_BYTE, 0, color[0]);
        glBindTexture(GL_TEXTURE_2D, height_map->tex[SNOW]);
        drawArray(dir);
    }
    //third pass
    glColorPointer(4, GL_UNSIGNED_BYTE, 0, color[1]);
    glBindTexture(GL_TEXTURE_2D, height_map->tex[ROCK]);
    drawArray(dir);

    glDisableClientState(GL_COLOR_ARRAY);
    //fourth pass
    if (tveVoyager::cv_posy->value.floating < tveVoyager::cv_water->value.floating)
    {
        glColor4f(1.0f, 1.0f, 1.0f, 0.1f);
        glBindTexture(GL_TEXTURE_2D, top->cur_water_tex);
        drawArray(dir);
    }

    glDepthMask(1);
    //FIXME: I don't know why I need to put that line before disabling TEXTURE_COORD in
    //       order to run perfectly on GeForce 2 Ultra under Linux with driver 0.9.767
    glClientActiveTextureARB(GL_TEXTURE1_ARB);
    if (top->lock) glUnlockArraysEXT();
    glDisableClientState(GL_TEXTURE_COORD_ARRAY);
    glDisableClientState(GL_VERTEX_ARRAY);
}
