/*
 * Copyright (c) 2001, Thierry Van Elsuw 
 * All rights reserved. 
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions 
 * are met:
 * 
 * Redistributions of source code must retain the above copyright 
 * notice, this list of conditions and the following disclaimer.
 * 
 * Redistributions in binary form must reproduce the above copyright 
 * notice, this list of conditions and the following disclaimer in the 
 * documentation and/or other materials provided with the distribution.
 * 
 * Neither the name of DS Improve nor the names of its contributors may 
 * be used to endorse or promote products derived from this software 
 * without specific prior written permission.
 * 
 * Module name   : tveQuadtree.cpp
 * Creation date : 23/02/2001
 * e-mail        : Thierry Van Elsuwe <tvelsuwe@swing.be>
 */

#include "voyager.h"
#include "tveQuadtree.h"


//=====================================================================
// tveQuadTree
//=====================================================================

/*
 *
 */
tveQuadTree::tveQuadTree(tveHeightMap *height_map)
{
    int j;

    this->height_map = height_map;
    sizex = height_map->height >> LEAF_GRANULARITY_SHIFT;
    sizey = height_map->width >> LEAF_GRANULARITY_SHIFT;
    pattern[TVE_VERTEX_ARRAY] = new tveVertexArray[sizex*sizey];
    tveVertexArray *array = (tveVertexArray *)pattern[TVE_VERTEX_ARRAY];
    leaves = new tveQuadLeaf*[sizex*sizey];
    //FIXME: ca fait mal de voir tant de memoire allouee
    for (j=0; j<sizex*sizey; j++)
    {
        array[j].vertices = new tveVertex[(LEAF_GRANULARITY+1)*(LEAF_GRANULARITY+1)];
        array[j].color[0] = new tveRGBA[(LEAF_GRANULARITY+1)*(LEAF_GRANULARITY+1)];
        array[j].color[1] = new tveRGBA[(LEAF_GRANULARITY+1)*(LEAF_GRANULARITY+1)];
        array[j].tex_coord[0] = new tveTexCoord[(LEAF_GRANULARITY+1)*(LEAF_GRANULARITY+1)];
        array[j].tex_coord[1] = new tveTexCoord[(LEAF_GRANULARITY+1)*(LEAF_GRANULARITY+1)];
        //normal, ...
    }

    // Create the vertex array
    for (j=0; j<sizey; j++)
    {
        int offset_z = j*LEAF_GRANULARITY;

        for (int i=0; i<sizex; i++)
        {
            int         offset_x    = i*LEAF_GRANULARITY;
            tveVertex   *vertices   = array[j*sizex+i].vertices;
            tveTexCoord *tex_coord[2] = {array[j*sizex+i].tex_coord[0], array[j*sizex+i].tex_coord[1]};
            tveRGBA     *color[2]   = {array[j*sizex+i].color[0], array[j*sizex+i].color[1]};

            array[j*sizex+i].height_map = height_map;
            array[j*sizex+i].top = this;
            for (int z=0; z<=LEAF_GRANULARITY; z++)
            {
                for (int x=0; x<=LEAF_GRANULARITY; x++)
                {
                    vertices[z*(LEAF_GRANULARITY+1)+x].x = (offset_x+x) * SCALE_X;
                    vertices[z*(LEAF_GRANULARITY+1)+x].y = height_map->detail[offset_x+x][offset_z+z] * SCALE_Y;
                    vertices[z*(LEAF_GRANULARITY+1)+x].z = (offset_z+z) * SCALE_Z;
                    tex_coord[0][z*(LEAF_GRANULARITY+1)+x].s = (offset_x+x)/(GLfloat)height_map->width;
                    tex_coord[0][z*(LEAF_GRANULARITY+1)+x].t = (offset_z+z)/(GLfloat)height_map->height;
                    tex_coord[1][z*(LEAF_GRANULARITY+1)+x].s = (GLfloat)x/2.0f;
                    tex_coord[1][z*(LEAF_GRANULARITY+1)+x].t = (GLfloat)z/2.0f;
                    color[0][z*(LEAF_GRANULARITY+1)+x][RED]   = 0xff;
                    color[0][z*(LEAF_GRANULARITY+1)+x][GREEN] = 0xff;
                    color[0][z*(LEAF_GRANULARITY+1)+x][BLUE]  = 0xff;
                    color[0][z*(LEAF_GRANULARITY+1)+x][ALPHA] = 0x00;
                    if (height_map->detail[offset_x+x][offset_z+z] > 128)
                    {
                        color[0][z*(LEAF_GRANULARITY+1)+x][ALPHA] = MIN((height_map->detail[offset_x+x][offset_z+z]-128)*2, 0xff);
                                                          //(height_map.detail[i][j]-height_map.detail[i][j])/20.0f;
                        //TODO: add snow at random position (must be done via GeoMap)
                    }
                    color[1][z*(LEAF_GRANULARITY+1)+x][RED]   = 0xff;
                    color[1][z*(LEAF_GRANULARITY+1)+x][GREEN] = 0xff;
                    color[1][z*(LEAF_GRANULARITY+1)+x][BLUE]  = 0xff;
                    color[1][z*(LEAF_GRANULARITY+1)+x][ALPHA] = 0x00;
                    {
                        int alpha =            (abs(height_map->detail[offset_x+x][offset_z+z]-height_map->detail[offset_x+x+1][offset_z+z])-8);
                            alpha = MAX(alpha, (abs(height_map->detail[offset_x+x][offset_z+z]-height_map->detail[offset_x+x][offset_z+z+1])-8));
                        if ((offset_x+x-1>=0) && (offset_z+z-1>=0))
                        {
                            alpha = MAX(alpha, (abs(height_map->detail[offset_x+x][offset_z+z]-height_map->detail[offset_x+x-1][offset_z+z])-8));
                            alpha = MAX(alpha, (abs(height_map->detail[offset_x+x][offset_z+z]-height_map->detail[offset_x+x][offset_z+z-1])-8));
                        }
                        if (alpha < 0) alpha = 0;
                        alpha *= 8;
                        color[1][z*(LEAF_GRANULARITY+1)+x][ALPHA] = MIN(alpha, 0xff);
                                                          //(height_map.detail[i][j]-height_map.detail[i][j])/20.0f;
                    }
                    // normal, ...
                }
            }
        }
    }
    array[0].faces_strip_detail = new GLuint[(LEAF_GRANULARITY+1)*LEAF_GRANULARITY*2];
    array[0].faces_strip_coarse = new GLuint[(LEAF_GRANULARITY/2+1)*(LEAF_GRANULARITY/2)*2];
    for (j=0; j<LEAF_GRANULARITY; j++)
    {
        int y_offset = j*(LEAF_GRANULARITY+1)*2;

        for (int i=0; i<=LEAF_GRANULARITY; i++)
        {
            array[0].faces_strip_detail[y_offset+i*2+0] = (j+0)*(LEAF_GRANULARITY+1)+i;
            array[0].faces_strip_detail[y_offset+i*2+1] = (j+1)*(LEAF_GRANULARITY+1)+i;
        }
    }
    for (j=0; j<LEAF_GRANULARITY/2; j++)
    {
        int y_offset = j*(LEAF_GRANULARITY/2+1)*2;

        for (int i=0; i<=LEAF_GRANULARITY/2; i++)
        {
            array[0].faces_strip_coarse[y_offset+i*2+0] = (j*2+0)*(LEAF_GRANULARITY+1)+i*2;
            array[0].faces_strip_coarse[y_offset+i*2+1] = (j*2+2)*(LEAF_GRANULARITY+1)+i*2;
        }
    }
    for (j=1; j<sizex*sizey; j++)
    {
        array[j].faces_strip_detail = array[0].faces_strip_detail;
        array[j].faces_strip_coarse = array[0].faces_strip_coarse;
    }

    // aabb_min.y & aabb_max.y will be computed while creating the quadtree
    tveVertex aabb_min, aabb_max;
    aabb_min.x = 0;
    aabb_min.y = -32000;
    aabb_min.z = 0;
    aabb_max.x = height_map->width;
    aabb_max.y = 32000;
    aabb_max.z = height_map->height;
    top = new tveQuadNode(this, &aabb_min, &aabb_max);
    build(top);
}


/*
 *
 */
tveQuadTree::~tveQuadTree()
{
    tveVertexArray *array = (tveVertexArray *)pattern[TVE_VERTEX_ARRAY];

    // delete the vertex array
    delete []array[0].faces_strip_detail;
    delete []array[0].faces_strip_coarse;
    for (int i=0; i<sizex*sizey; i++)
    {
        delete []array[i].tex_coord[0];
        delete []array[i].tex_coord[1];
        delete []array[i].color[0];
        delete []array[i].color[1];
        delete []array[i].vertices;
        //normal, ...
    }
    delete []array;

    delete []leaves;
    delete top;
}


/*
 *
 */
void tveQuadTree::build(tveQuadNode *node)
{
    tveVertex aabb_min, aabb_max;

    // continue if the current sector is not small enough
    if ((node->aabb_max.x-node->aabb_min.x) > (LEAF_GRANULARITY+1)*2)
    {
        aabb_min.x = node->aabb_min.x;
        aabb_min.y = node->aabb_min.y;
        aabb_min.z = (node->aabb_max.z+node->aabb_min.z)/2;
        aabb_max.x = (node->aabb_max.x+node->aabb_min.x)/2;
        aabb_max.y = node->aabb_max.y;
        aabb_max.z = node->aabb_max.z;
        node->upper_left   = new tveQuadNode(this, &aabb_min, &aabb_max);
        build((tveQuadNode*)node->upper_left);

        aabb_min.x = (node->aabb_max.x+node->aabb_min.x)/2;
        aabb_min.y = node->aabb_min.y;
        aabb_min.z = (node->aabb_max.z+node->aabb_min.z)/2;
        aabb_max.x = node->aabb_max.x;
        aabb_max.y = node->aabb_max.y;
        aabb_max.z = node->aabb_max.z;
        node->upper_right  = new tveQuadNode(this, &aabb_min, &aabb_max);
        build((tveQuadNode*)node->upper_right);

        aabb_min.x = node->aabb_min.x;
        aabb_min.y = node->aabb_min.y;
        aabb_min.z = node->aabb_min.z;
        aabb_max.x = (node->aabb_max.x+node->aabb_min.x)/2;
        aabb_max.y = node->aabb_max.y;
        aabb_max.z = (node->aabb_max.z+node->aabb_min.z)/2;
        node->lower_left  = new tveQuadNode(this, &aabb_min, &aabb_max);
        build((tveQuadNode*)node->lower_left);

        aabb_min.x = (node->aabb_max.x+node->aabb_min.x)/2;
        aabb_min.y = node->aabb_min.y;
        aabb_min.z = node->aabb_min.z;
        aabb_max.x = node->aabb_max.x;
        aabb_max.y = node->aabb_max.y;
        aabb_max.z = (node->aabb_max.z+node->aabb_min.z)/2;
        node->lower_right = new tveQuadNode(this, &aabb_min, &aabb_max);
        build((tveQuadNode*)node->lower_right);

        return;
    }

    // Now, we are going to create the leaves
    aabb_min.x = node->aabb_min.x;
    aabb_min.y = node->aabb_min.y;
    aabb_min.z = (node->aabb_max.z+node->aabb_min.z)/2;
    aabb_max.x = (node->aabb_max.x+node->aabb_min.x)/2;
    aabb_max.y = node->aabb_max.y;
    aabb_max.z = node->aabb_max.z;
    node->upper_left   = new tveQuadLeaf(this, &aabb_min, &aabb_max);

    aabb_min.x = (node->aabb_max.x+node->aabb_min.x)/2;
    aabb_min.y = node->aabb_min.y;
    aabb_min.z = (node->aabb_max.z+node->aabb_min.z)/2;
    aabb_max.x = node->aabb_max.x;
    aabb_max.y = node->aabb_max.y;
    aabb_max.z = node->aabb_max.z;
    node->upper_right  = new tveQuadLeaf(this, &aabb_min, &aabb_max);

    aabb_min.x = node->aabb_min.x;
    aabb_min.y = node->aabb_min.y;
    aabb_min.z = node->aabb_min.z;
    aabb_max.x = (node->aabb_max.x+node->aabb_min.x)/2;
    aabb_max.y = node->aabb_max.y;
    aabb_max.z = (node->aabb_max.z+node->aabb_min.z)/2;
    node->lower_left  = new tveQuadLeaf(this, &aabb_min, &aabb_max);

    aabb_min.x = (node->aabb_max.x+node->aabb_min.x)/2;
    aabb_min.y = node->aabb_min.y;
    aabb_min.z = node->aabb_min.z;
    aabb_max.x = node->aabb_max.x;
    aabb_max.y = node->aabb_max.y;
    aabb_max.z = (node->aabb_max.z+node->aabb_min.z)/2;
    node->lower_right = new tveQuadLeaf(this, &aabb_min, &aabb_max);
}


/*
 *
 */
void tveQuadTree::addLeaf(tveQuadLeaf *leaf)
{
    leaf->num = ((int)leaf->aabb_min.x)/LEAF_GRANULARITY + (((int)leaf->aabb_min.z)/LEAF_GRANULARITY)*sizex;
    leaves[leaf->num] = leaf;
}


/*
 *
 */
tveQuadLeaf *tveQuadTree::getNeighbour(tveQuadLeaf *leaf, int dir)
{
    switch (dir)
    {
        case DIR_NORTH:
            if (leaf->num >= sizex)
            {
                return leaves[leaf->num-sizex];
            }
            break;
        case DIR_SOUTH:
            if (leaf->num < sizex*(sizey-1))
            {
                return leaves[leaf->num+sizex];
            }
            break;
        case DIR_WEST:
            if ( (leaf->num % sizex) != 0 )
            {
                return leaves[leaf->num-1];
            }
            break;
        case DIR_EAST:
            if ( ((leaf->num+1) % sizex) != 0 )
            {
                return leaves[leaf->num+1];
            }
            break;
        default:
            return leaves[leaf->num];
    }
    return NULL;
}


/*
 *
 */
int tveQuadTree::getLOD(tveQuadLeaf *leaf)
{
    tveQuadLeaf *neighbour;
    int         res = 0;
    int         exp = 1;

    for (int i=0; i<NUM_DIRECTION; i++)
    {
        if ( (neighbour = getNeighbour(leaf, i)) )
        {
            res += ( leaf->getDetail() != neighbour->getDetail() ) ?  exp : 0;
        }
        exp *= 2;
    }
    return res;
}


/*
 *
 */
tveVertexArray *tveQuadTree::getVertexArray(tveVertex *aabb_min, tveVertex *aabb_max)
{
    int idx = (((int)aabb_min->x)>>LEAF_GRANULARITY_SHIFT)+((sizex*((int)aabb_min->z))>>LEAF_GRANULARITY_SHIFT);
    return &((tveVertexArray *)pattern[TVE_VERTEX_ARRAY])[idx];
}


/*
 *
 */
//TODO: use intersect/reject/accept to know if we can accept all leaves in a bbox
void tveQuadTree::getFrustum()
{
   GLfloat proj[16];
   GLfloat modl[16];
   GLfloat clip[16];
   GLfloat t;

   glGetFloatv(GL_PROJECTION_MATRIX, proj);
   glGetFloatv(GL_MODELVIEW_MATRIX, modl);

   // Combine the two matrices (multiply projection by modelview)
   clip[ 0] = modl[ 0]*proj[ 0] + modl[ 1]*proj[ 4] + modl[ 2]*proj[ 8] + modl[ 3]*proj[12];
   clip[ 1] = modl[ 0]*proj[ 1] + modl[ 1]*proj[ 5] + modl[ 2]*proj[ 9] + modl[ 3]*proj[13];
   clip[ 2] = modl[ 0]*proj[ 2] + modl[ 1]*proj[ 6] + modl[ 2]*proj[10] + modl[ 3]*proj[14];
   clip[ 3] = modl[ 0]*proj[ 3] + modl[ 1]*proj[ 7] + modl[ 2]*proj[11] + modl[ 3]*proj[15];

   clip[ 4] = modl[ 4]*proj[ 0] + modl[ 5]*proj[ 4] + modl[ 6]*proj[ 8] + modl[ 7]*proj[12];
   clip[ 5] = modl[ 4]*proj[ 1] + modl[ 5]*proj[ 5] + modl[ 6]*proj[ 9] + modl[ 7]*proj[13];
   clip[ 6] = modl[ 4]*proj[ 2] + modl[ 5]*proj[ 6] + modl[ 6]*proj[10] + modl[ 7]*proj[14];
   clip[ 7] = modl[ 4]*proj[ 3] + modl[ 5]*proj[ 7] + modl[ 6]*proj[11] + modl[ 7]*proj[15];

   clip[ 8] = modl[ 8]*proj[ 0] + modl[ 9]*proj[ 4] + modl[10]*proj[ 8] + modl[11]*proj[12];
   clip[ 9] = modl[ 8]*proj[ 1] + modl[ 9]*proj[ 5] + modl[10]*proj[ 9] + modl[11]*proj[13];
   clip[10] = modl[ 8]*proj[ 2] + modl[ 9]*proj[ 6] + modl[10]*proj[10] + modl[11]*proj[14];
   clip[11] = modl[ 8]*proj[ 3] + modl[ 9]*proj[ 7] + modl[10]*proj[11] + modl[11]*proj[15];

   clip[12] = modl[12]*proj[ 0] + modl[13]*proj[ 4] + modl[14]*proj[ 8] + modl[15]*proj[12];
   clip[13] = modl[12]*proj[ 1] + modl[13]*proj[ 5] + modl[14]*proj[ 9] + modl[15]*proj[13];
   clip[14] = modl[12]*proj[ 2] + modl[13]*proj[ 6] + modl[14]*proj[10] + modl[15]*proj[14];
   clip[15] = modl[12]*proj[ 3] + modl[13]*proj[ 7] + modl[14]*proj[11] + modl[15]*proj[15];

   // Extract the numbers for the RIGHT plane
   frustum[0][0] = clip[ 3] - clip[ 0];
   frustum[0][1] = clip[ 7] - clip[ 4];
   frustum[0][2] = clip[11] - clip[ 8];
   frustum[0][3] = clip[15] - clip[12];
   // Normalize the result
   t = 1.0f/(GLfloat)sqrt(frustum[0][0]*frustum[0][0] + frustum[0][1]*frustum[0][1] + frustum[0][2]*frustum[0][2]);
   frustum[0][0] *= t;
   frustum[0][1] *= t;
   frustum[0][2] *= t;
   frustum[0][3] *= t;

   // Extract the numbers for the LEFT plane
   frustum[1][0] = clip[ 3] + clip[ 0];
   frustum[1][1] = clip[ 7] + clip[ 4];
   frustum[1][2] = clip[11] + clip[ 8];
   frustum[1][3] = clip[15] + clip[12];
   // Normalize the result
   t = 1.0f/(GLfloat)sqrt(frustum[1][0]*frustum[1][0] + frustum[1][1]*frustum[1][1] + frustum[1][2]*frustum[1][2]);
   frustum[1][0] *= t;
   frustum[1][1] *= t;
   frustum[1][2] *= t;
   frustum[1][3] *= t;

   // Extract the BOTTOM plane
   frustum[2][0] = clip[ 3] + clip[ 1];
   frustum[2][1] = clip[ 7] + clip[ 5];
   frustum[2][2] = clip[11] + clip[ 9];
   frustum[2][3] = clip[15] + clip[13];
   // Normalize the result
   t = 1.0f/(GLfloat)sqrt(frustum[2][0]*frustum[2][0] + frustum[2][1]*frustum[2][1] + frustum[2][2]*frustum[2][2]);
   frustum[2][0] *= t;
   frustum[2][1] *= t;
   frustum[2][2] *= t;
   frustum[2][3] *= t;

   // Extract the TOP plane
   frustum[3][0] = clip[ 3] - clip[ 1];
   frustum[3][1] = clip[ 7] - clip[ 5];
   frustum[3][2] = clip[11] - clip[ 9];
   frustum[3][3] = clip[15] - clip[13];
   // Normalize the result
   t = 1.0f/(GLfloat)sqrt(frustum[3][0]*frustum[3][0] + frustum[3][1]*frustum[3][1] + frustum[3][2]*frustum[3][2]);
   frustum[3][0] *= t;
   frustum[3][1] *= t;
   frustum[3][2] *= t;
   frustum[3][3] *= t;

   // Extract the FAR plane
   frustum[4][0] = clip[ 3] - clip[ 2];
   frustum[4][1] = clip[ 7] - clip[ 6];
   frustum[4][2] = clip[11] - clip[10];
   frustum[4][3] = clip[15] - clip[14];
   // Normalize the result
   t = 1.0f/(GLfloat)sqrt(frustum[4][0]*frustum[4][0] + frustum[4][1]*frustum[4][1] + frustum[4][2]*frustum[4][2]);
   frustum[4][0] *= t;
   frustum[4][1] *= t;
   frustum[4][2] *= t;
   frustum[4][3] *= t;

   // Extract the NEAR plane
   frustum[5][0] = clip[ 3] + clip[ 2];
   frustum[5][1] = clip[ 7] + clip[ 6];
   frustum[5][2] = clip[11] + clip[10];
   frustum[5][3] = clip[15] + clip[14];
   // Normalize the result
   t = 1.0f/(GLfloat)sqrt(frustum[5][0]*frustum[5][0] + frustum[5][1]*frustum[5][1] + frustum[5][2]*frustum[5][2]);
   frustum[5][0] *= t;
   frustum[5][1] *= t;
   frustum[5][2] *= t;
   frustum[5][3] *= t;
}


/*
 *
 */
int tveQuadTree::cubeInFrustum(tveQuadSector *sector)
{
    tveVertex aabb_min;
    tveVertex aabb_max;

    aabb_min.x = sector->aabb_min.x*SCALE_X;
    aabb_min.y = sector->aabb_min.y*SCALE_Y;
    aabb_min.z = sector->aabb_min.z*SCALE_Z;
    aabb_max.x = sector->aabb_max.x*SCALE_X;
    aabb_max.y = sector->aabb_max.y*SCALE_Y;
    aabb_max.z = sector->aabb_max.z*SCALE_Z;
    for (int p=0; p<6; p++)
    {
        if ((frustum[p][0]*aabb_min.x + frustum[p][1]*aabb_min.y + frustum[p][2]*aabb_min.z + frustum[p][3]) > 0.0f)
            continue;
        if ((frustum[p][0]*aabb_max.x + frustum[p][1]*aabb_min.y + frustum[p][2]*aabb_min.z + frustum[p][3]) > 0.0f)
            continue;
        if ((frustum[p][0]*aabb_min.x + frustum[p][1]*aabb_max.y + frustum[p][2]*aabb_min.z + frustum[p][3]) > 0.0f)
            continue;
        if ((frustum[p][0]*aabb_max.x + frustum[p][1]*aabb_max.y + frustum[p][2]*aabb_min.z + frustum[p][3]) > 0.0f)
            continue;
        if ((frustum[p][0]*aabb_min.x + frustum[p][1]*aabb_min.y + frustum[p][2]*aabb_max.z + frustum[p][3]) > 0.0f)
            continue;
        if ((frustum[p][0]*aabb_max.x + frustum[p][1]*aabb_min.y + frustum[p][2]*aabb_max.z + frustum[p][3]) > 0.0f)
            continue;
        if ((frustum[p][0]*aabb_min.x + frustum[p][1]*aabb_max.y + frustum[p][2]*aabb_max.z + frustum[p][3]) > 0.0f)
            continue;
        if ((frustum[p][0]*aabb_max.x + frustum[p][1]*aabb_max.y + frustum[p][2]*aabb_max.z + frustum[p][3]) > 0.0f)
            continue;
        return 0;
    }
    return 1;
}


/*
 *
 */
/*
int tveQuadTree::cubeInFrustum(tveVertex *aabb_min, tveVertex *aabb_max)
{
    int c, c2 = 0;

    for (int p=0; p<6; p++)
    {
        c = 0;
        if ((frustum[p][0]*aabb_min->x + frustum[p][1]*aabb_min->y + frustum[p][2]*aabb_min->z + frustum[p][3]) > 0.0f)
            c++;
        if ((frustum[p][0]*aabb_max->x + frustum[p][1]*aabb_min->y + frustum[p][2]*aabb_min->z + frustum[p][3]) > 0.0f)
            c++;
        if ((frustum[p][0]*aabb_min->x + frustum[p][1]*aabb_max->y + frustum[p][2]*aabb_min->z + frustum[p][3]) > 0.0f)
            c++;
        if ((frustum[p][0]*aabb_max->x + frustum[p][1]*aabb_max->y + frustum[p][2]*aabb_min->z + frustum[p][3]) > 0.0f)
            c++;
        if ((frustum[p][0]*aabb_min->x + frustum[p][1]*aabb_min->y + frustum[p][2]*aabb_max->z + frustum[p][3]) > 0.0f)
            c++;
        if ((frustum[p][0]*aabb_max->x + frustum[p][1]*aabb_min->y + frustum[p][2]*aabb_max->z + frustum[p][3]) > 0.0f)
            c++;
        if ((frustum[p][0]*aabb_min->x + frustum[p][1]*aabb_max->y + frustum[p][2]*aabb_max->z + frustum[p][3]) > 0.0f)
            c++;
        if ((frustum[p][0]*aabb_max->x + frustum[p][1]*aabb_max->y + frustum[p][2]*aabb_max->z + frustum[p][3]) > 0.0f)
            c++;
        if (c==0)
            return 0;
        if (c==8)
            c2++;
    }
    return (c2 == 6) ? 2 : 1;
}
*/


/*
 *
 */
float tveQuadTree::distp2p(tveVertex *p1, tveVertex *p2)
{
    float x = p1->x - p2->x;
    float y = p1->y - p2->y;
    float z = p1->z - p2->z;
    return x*x + y*y + z*z;
}


/*
 *
 */
void tveQuadTree::renderGeoMap(bool lock)
{
    top->renderGeoMap();
}


/*
 *
 */
void tveQuadTree::renderWater()
{
    static float num = 0.0f;

    num += 0.02f * tveVoyager::ratio;
    if (num > 16.0f)
    {
        num -= 16.0f;
    }
    glActiveTextureARB(GL_TEXTURE1_ARB);
    glEnable(GL_TEXTURE_2D);
    glBindTexture(GL_TEXTURE_2D, height_map->tex[SKY]);
    glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_DECAL);
    glActiveTextureARB(GL_TEXTURE0_ARB);
    glEnable(GL_TEXTURE_2D);
    cur_water_tex = height_map->tex[WATER+(int)num];
    glBindTexture(GL_TEXTURE_2D, cur_water_tex);
    glColor4f(1.0f, 1.0f, 1.0f, 0.85f);
    top->renderWater();
    glActiveTextureARB(GL_TEXTURE1_ARB);
    glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
}


/*
 *
 */
void tveQuadTree::renderTerrain(bool lock)
{
    getFrustum();
    num_sect = 0;
    num_poly = 0;
    this->lock = lock;
    
#ifndef DEMO_DS
    static int old = SDL_GetTicks(); // profiling stuffs
    int current;
    
    current = SDL_GetTicks();
    console->printf(10, 200, 0xffFFffFF, 0.5f, "idle   : %d", current-old);
    old = current;
#endif
    top->preCalculate();

#ifndef DEMO_DS
    current = SDL_GetTicks();
    console->printf(10, 180, 0xffFFffFF, 0.5f, "precalc: %d", current-old);
    old = current;
#endif

    top->renderTerrain();

#ifndef DEMO_DS
    current = SDL_GetTicks();
    console->printf(10, 160, 0xffFFffFF, 0.5f, "render : %d", current-old);
    old = current;

    console->printf(10, (int)tveVoyager::cv_screenheight->value.integer-40, 0xffFFffFF, 0.5f, "Num sect: %d (%d)", num_sect, num_poly);
#endif
}

