/*
 * Copyright (c) 2001, Thierry Van Elsuw 
 * All rights reserved. 
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions 
 * are met:
 * 
 * Redistributions of source code must retain the above copyright 
 * notice, this list of conditions and the following disclaimer.
 * 
 * Redistributions in binary form must reproduce the above copyright 
 * notice, this list of conditions and the following disclaimer in the 
 * documentation and/or other materials provided with the distribution.
 * 
 * Neither the name of DS Improve nor the names of its contributors may 
 * be used to endorse or promote products derived from this software 
 * without specific prior written permission.
 * 
 * Module name   : tveQuadtree.h
 * Creation date : 23/02/2001
 * e-mail        : Thierry Van Elsuwe <tvelsuwe@swing.be>
 */

#ifndef _TVEQUADTREE_H_
#define _TVEQUADTREE_H_

#include "voyager.h"
#include "tvePattern.h"
#include "tveHeightmap.h"
#include "tveSector.h"


class tveQuadTree;
class tveQuadSector;
class tveQuadNode;
class tveQuadLeaf;


enum tveDirection
{
    DIR_NORTH = 0,
    DIR_SOUTH,
    DIR_WEST,
    DIR_EAST,
    NUM_DIRECTION
};


enum tveDirectionMask
{
    DIR_NORTH_MASK = 0x01,
    DIR_SOUTH_MASK = 0x02,
    DIR_WEST_MASK  = 0x04,
    DIR_EAST_MASK  = 0x08,
};


class tveQuadTree
{
private:
    tveQuadNode     *top;
    tveHeightMap    *height_map;
    Uint16          sizex;
    Uint16          sizey;
    tveQuadLeaf     **leaves; // used for accessing the leaves without going through the quad-tree
    tvePattern      *pattern[TVE_NUM_RENDER_METHOD];
    GLfloat         frustum[6][4];

    void  getFrustum();
    tveQuadLeaf *getNeighbour(tveQuadLeaf *leaf, int dir);
public:
    bool    lock;
    char    pass;
    float   wind;
    GLuint  cur_water_tex;
    int     num_sect;
    int     num_poly;

    tveQuadTree(tveHeightMap *height_map);
    ~tveQuadTree();
    void  addLeaf(tveQuadLeaf *leaf);
    int   getLOD(tveQuadLeaf *leaf);
    void  renderTerrain(bool lock);
    void  renderGeoMap(bool lock);
    void  renderWater();
    void  build(tveQuadNode *node);
    int   cubeInFrustum(tveQuadSector *node);
    float distp2p(tveVertex *p1, tveVertex *p2);
    tveVertexArray *getVertexArray(tveVertex *aabb_min, tveVertex *aabb_max);
};


#endif
