/*
 * Copyright (c) 2001, Thierry Van Elsuw 
 * All rights reserved. 
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions 
 * are met:
 * 
 * Redistributions of source code must retain the above copyright 
 * notice, this list of conditions and the following disclaimer.
 * 
 * Redistributions in binary form must reproduce the above copyright 
 * notice, this list of conditions and the following disclaimer in the 
 * documentation and/or other materials provided with the distribution.
 * 
 * Neither the name of DS Improve nor the names of its contributors may 
 * be used to endorse or promote products derived from this software 
 * without specific prior written permission.
 * 
 * Module name   : tveSector.h
 * Creation date : 26/04/2001
 * e-mail        : Thierry Van Elsuwe <tvelsuwe@swing.be>
 */

#ifndef _TVESECTOR_H_
#define _TVESECTOR_H_

#include "voyager.h"
#include "tveQuadtree.h"
#include "tvePattern.h"


class tveQuadTree;
class tveQuadSector;
class tveQuadNode;
class tveQuadLeaf;


class tveQuadSector
{
protected:
    tveQuadTree *top;
    bool        visible;
public:
    tveVertex   aabb_min;
    tveVertex   aabb_max;
    tveVertex   aabb_center;
    tveQuadSector(tveVertex *aabb_min, tveVertex *aabb_max);
    virtual ~tveQuadSector();
    virtual void preCalculate()  = 0;
    virtual void renderTerrain() = 0;
    virtual void renderGeoMap()  = 0;
    virtual void renderWater()   = 0;
};


class tveQuadNode : public tveQuadSector
{
public:
    tveQuadSector *upper_left;
    tveQuadSector *upper_right;
    tveQuadSector *lower_left;
    tveQuadSector *lower_right;
    tveQuadNode(tveQuadTree *top, tveVertex *aabb_min, tveVertex *aabb_max);
    ~tveQuadNode();
    void preCalculate();
    void renderTerrain();
    void renderGeoMap();
    void renderWater();
};


class tveQuadLeaf : public tveQuadSector
{
private:
    Sint16                  *detail2coarse;
    Uint8                   *coarse_color;
    tveVertexArray          *array;
public:
    int num; // we attribute a number to this leaf for quick access its neighbours
    tveQuadLeaf(tveQuadTree *top, tveVertex *aabb_min, tveVertex *aabb_max);
    ~tveQuadLeaf();
    void preCalculate();
    void renderTerrain();
    void renderGeoMap();
    void renderWater();
    bool getDetail();
};

#endif
