/*
 * Copyright (c) 2001, Thierry Van Elsuw 
 * All rights reserved. 
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions 
 * are met:
 * 
 * Redistributions of source code must retain the above copyright 
 * notice, this list of conditions and the following disclaimer.
 * 
 * Redistributions in binary form must reproduce the above copyright 
 * notice, this list of conditions and the following disclaimer in the 
 * documentation and/or other materials provided with the distribution.
 * 
 * Neither the name of DS Improve nor the names of its contributors may 
 * be used to endorse or promote products derived from this software 
 * without specific prior written permission.
 * 
 * Module name   : tveTexture.cpp
 * Creation date : 16/02/2001
 * e-mail        : Thierry Van Elsuwe <tvelsuwe@swing.be>
 */

#include "voyager.h"
#include "tveTexture.h"

char    *tveTexture::tex_name[MAX_TEXTURE] = {NULL};
GLuint  tveTexture::tex_num[MAX_TEXTURE];


/*
 *
 */
void tveTexture::init()
{
    for (int i=0; i<MAX_TEXTURE; i++)
    {
        tex_name[i] = NULL;
    }
}


/*
 *
 */
void tveTexture::saveTGA(char *filename, int width, int height, tveRGB *buffer)
{
    int          fd;
    long         size, i;
    tveTGAHeader tga_hdr;
    char         tmp;

    if (sizeof(tveTGAHeader) != 18) 
    {
        printf("Erreur: saveTGA(): sizeof(tveTGAHeader) = %u\n", sizeof(tveTGAHeader));
        exit(-1);
    }
    console->printf("saving... %s\n", filename);
    memset(&tga_hdr, 0, sizeof(tga_hdr));
    tga_hdr.width = width;
    tga_hdr.height = height;
    tga_hdr.image_pix_size = 24;
    tga_hdr.image_type = 2;
    tga_hdr.image_descriptor = 32;
    size = (long)width * (long)height * 3L;
    // swap colours
    for (i=0; i<tga_hdr.width; i++)
    {
        for (int j=0; j<tga_hdr.height; j++) 
        {
            tmp = buffer[tga_hdr.width*j+i][RED];
            buffer[tga_hdr.width*j+i][RED] = buffer[tga_hdr.width*j+i][BLUE];
            buffer[tga_hdr.width*j+i][BLUE] = tmp;
        }
    }
    // swap lines
    for (i=0; i<tga_hdr.width; i++)
    {
        for (int j=0; j<tga_hdr.height/2; j++)
        {
            for (int k=0; k<3; k++) 
            {
                tmp = buffer[tga_hdr.width*j+i][k];
                buffer[tga_hdr.width*j+i][k] = buffer[tga_hdr.width*(tga_hdr.height-j-1)+i][k];
                buffer[tga_hdr.width*(tga_hdr.height-j-1)+i][k] = tmp;
            }
        }
    }
    char file_name[256];
    sprintf(file_name, "%s%s", tveVoyager::path, filename);
    fd = open(file_name, O_WRONLY|O_BINARY|O_CREAT, 0644);
    if (fd < 0)
    {
        console->printf("cannot save to %s ", filename);
        return;
    }
    write(fd, &tga_hdr, sizeof(tveTGAHeader));
    write(fd, buffer, size);
    close(fd);
}


/*
 *
 */
//TODO: manage endianess
GLuint tveTexture::load(char *filename)
{
    gzFile          f;
    long            i, num;
    tveTGAHeader    tga_hdr;
    tveRGB          *image;  // void * (ou char*) si type est utilis
    char            tmp;
    char            *field;
    char            file_name[256];

    //search if we have already bound that texture
    for (num=0; tex_name[num]; num++)
    {
        if (!strcmp(tex_name[num], filename))
        {
            return tex_num[num];
        }
    }
    tex_name[num] = strdup(filename);

    sprintf(file_name, "../data/textures/%s", filename);
    if ( !(f = tveVoyager::openFile(file_name, "rb")) )
    {
        console->printf("cannot load texture %s ", filename);
        return 0;
    }
    if (sizeof(tveTGAHeader) != 18)
    {
        printf("Erreur: loadTexture(): sizeof(TGAHeader) = %u\n", sizeof(tveTGAHeader));
        exit(-1);
    }
    gzread(f, &tga_hdr, sizeof(tveTGAHeader));
    tga_hdr.width  = SDL_SwapLE16(tga_hdr.width);
    tga_hdr.height = SDL_SwapLE16(tga_hdr.height);
    printf("loading... %-24s  | w: %d  h: %d  bbp: %d\n",
            filename, tga_hdr.width, tga_hdr.height, tga_hdr.image_pix_size);
    field = new char[tga_hdr.id_field_length];
    gzread(f, field, tga_hdr.id_field_length); // ignore les informations complmentaires
    delete []field;
    image = new tveRGB[tga_hdr.width*tga_hdr.height];
    gzread(f, image, (long)tga_hdr.width * (long)tga_hdr.height * (long)sizeof(tveRGB));
    gzclose(f);
    // swap colours
    for (i=0; i<tga_hdr.width; i++)
    {
        for (int j=0; j<tga_hdr.height; j++)
        {
            tmp = image[tga_hdr.width*j+i][RED];
            image[tga_hdr.width*j+i][RED] = image[tga_hdr.width*j+i][BLUE];
            image[tga_hdr.width*j+i][BLUE] = tmp;
        }
    }
    /* // swap lines
    for (i=0; i<tga_hdr.width; i++)
    {
        for (int j=0; j<tga_hdr.height/2; j++)
        {
            for (int k=0; k<3; k++)
            {
                tmp = image[tga_hdr.width*j+i][k];
                image[tga_hdr.width*j+i][k] = image[tga_hdr.width*(tga_hdr.height-j-1)+i][k];
                image[tga_hdr.width*(tga_hdr.height-j-1)+i][k] = tmp;
            }
        }
    } */
    glGenTextures(1, &tex_num[num]);
    glBindTexture(GL_TEXTURE_2D, tex_num[num]);
    gluBuild2DMipmaps(GL_TEXTURE_2D, GL_RGB, tga_hdr.width, tga_hdr.height, GL_RGB, GL_UNSIGNED_BYTE, image);
    delete []image;
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_NEAREST);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR_MIPMAP_NEAREST);

    return tex_num[num];
}


/*
 *
 */
//TODO: manage endianess
GLuint tveTexture::loada(char *filename, tveRGBA color)
{
    gzFile          f;
    long            i, num;
    tveTGAHeader    tga_hdr;
    tveRGB          *image_rgb;
    tveRGBA         *image_rgba;
    char            tmp;
    char            *field;
    char            file_name[256];

    //search if we have already bound that texture
    for (num=0; tex_name[num]; num++)
    {
        if (!strcmp(tex_name[num], filename))
        {
            return tex_num[num];
        }
    }
    tex_name[num] = strdup(filename);

    sprintf(file_name, "../data/textures/%s", filename);
    if ( !(f = tveVoyager::openFile(file_name, "rb")) )
    {
        console->printf("cannot load texture %s ", filename);
        return 0;
    }
    if (sizeof(tveTGAHeader) != 18)
    {
        printf("Erreur: loadTexture(): sizeof(TGAHeader) = %u\n", sizeof(tveTGAHeader));
        exit(-1);
    }
    gzread(f, &tga_hdr, sizeof(tveTGAHeader));
    tga_hdr.width  = SDL_SwapLE16(tga_hdr.width);
    tga_hdr.height = SDL_SwapLE16(tga_hdr.height);
    printf("loading... %-24s  | w: %d  h: %d  bbp: %d\n",
            filename, tga_hdr.width, tga_hdr.height, tga_hdr.image_pix_size);
    field = new char[tga_hdr.id_field_length];
    gzread(f, field, tga_hdr.id_field_length); // ignore les informations complmentaires
    delete []field;
    image_rgb  = new tveRGB[tga_hdr.width*tga_hdr.height];
    image_rgba = new tveRGBA[tga_hdr.width*tga_hdr.height];
    gzread(f, image_rgb, (long)tga_hdr.width * (long)tga_hdr.height * (long)sizeof(tveRGB));
    gzclose(f);
    // swap colours
    for (i=0; i<tga_hdr.width; i++)
    {
        for (int j=0; j<tga_hdr.height; j++)
        {
            tmp = image_rgb[tga_hdr.width*j+i][RED];
            image_rgb[tga_hdr.width*j+i][RED] = image_rgb[tga_hdr.width*j+i][BLUE];
            image_rgb[tga_hdr.width*j+i][BLUE] = tmp;
        }
    }
    for (i=0; i<tga_hdr.width; i++)
    {
        for (int j=0; j<tga_hdr.height; j++)
        {
            //if ( (image_rgb[tga_hdr.width*j+i][RED]  == color[RED])   &&
            //     (image_rgb[tga_hdr.width*j+i][GREEN]== color[GREEN]) &&
            //     (image_rgb[tga_hdr.width*j+i][BLUE] == color[BLUE]) )
            if ((image_rgb[tga_hdr.width*j+i][RED] > 128) || (image_rgb[tga_hdr.width*j+i][BLUE] > 128))
            {
                image_rgba[tga_hdr.width*j+i][RED]   = 0;
                image_rgba[tga_hdr.width*j+i][GREEN] = 0;
                image_rgba[tga_hdr.width*j+i][BLUE]  = 0;
                image_rgba[tga_hdr.width*j+i][ALPHA] = 0;
            }
            else
            {
                image_rgba[tga_hdr.width*j+i][RED]   = image_rgb[tga_hdr.width*j+i][RED];
                image_rgba[tga_hdr.width*j+i][GREEN] = image_rgb[tga_hdr.width*j+i][GREEN];
                image_rgba[tga_hdr.width*j+i][BLUE]  = image_rgb[tga_hdr.width*j+i][BLUE];
                image_rgba[tga_hdr.width*j+i][ALPHA] = 255;
            }
        }
    }
    glGenTextures(1, &tex_num[num]);
    glBindTexture(GL_TEXTURE_2D, tex_num[num]);
    gluBuild2DMipmaps(GL_TEXTURE_2D, GL_RGBA, tga_hdr.width, tga_hdr.height, GL_RGBA, GL_UNSIGNED_BYTE, image_rgba);
    delete []image_rgb;
    delete []image_rgba;
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_NEAREST);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR_MIPMAP_NEAREST);

    return tex_num[num];
}


/*
 *
 */
//TODO: manage endianess
GLuint tveTexture::loadPCX(char *filename)
{
    tvePCXHeader header;
    char    file_name[256];
    tveRGB  palette[256];
    tveRGB  *image;
    int     w, h;
    int     ptr = 0;
    int     ch, rep;
    int     size;
    long    num;
    gzFile  f;

    //search if we have already bound that texture
    for (num=0; tex_name[num]; num++)
    {
        if (!strcmp(tex_name[num], filename))
        {
            return tex_num[num];
        }
    }
    tex_name[num] = strdup(filename);

    //TODO: manage that better!!!
    {
        int i;

        sprintf(file_name, "../data/models/");
        int pos = strlen(file_name);
        for (i=pos; (*filename) != '\0'; i++, filename++)
        {
            if ((*filename == '\\') || (*filename == '/'))
                i = pos-1;
            else
                file_name[i] = *filename;
        }
        file_name[i] = '\0';
    }

    //sprintf(file_name, "../data/models/%s", filename);
    if ( !(f = tveVoyager::openFile(file_name, "rb")) )
    {
        console->printf("cannot load texture %s ", filename);
        return 0;
    }
    if (gzread(f, &header, sizeof(tvePCXHeader)) < 0)
    {
        gzclose(f);
        return 0;
    }
    header.xmax  = SDL_SwapLE16(header.xmax);
    header.ymax  = SDL_SwapLE16(header.ymax);
    header.xmin  = SDL_SwapLE16(header.xmin);
    header.ymin  = SDL_SwapLE16(header.ymin);

    //gzseek(f, -768, SEEK_END);
    //FIXME
    while (!gzeof(f))
    {
        ch = gzgetc(f);
    }
    gzseek(f, -768, SEEK_CUR);

    if (gzread(f, palette, 768) < 0)
    {
        gzclose(f);
        return 0;
    }
    w = header.xmax - header.xmin + 1;
    h = header.ymax - header.ymin + 1;
    size = w * h;
    image = new tveRGB[size];

    gzseek(f, sizeof(tvePCXHeader), SEEK_SET);
    while (ptr < size)
    {
        ch = gzgetc(f);
        if (ch >= 192)
        {
            rep = ch - 192;
            ch = gzgetc(f);
        }
        else 
        {
            rep = 1;
        }
        while (rep--)
        {
            image[ptr][RED]   = palette[ch][RED];
            image[ptr][GREEN] = palette[ch][GREEN];
            image[ptr][BLUE]  = palette[ch][BLUE];
            ptr++;
        }   
    }
    gzclose(f);

    glGenTextures(1, &tex_num[num]);
    glBindTexture(GL_TEXTURE_2D, tex_num[num]);
    gluBuild2DMipmaps(GL_TEXTURE_2D, GL_RGB, w, h, GL_RGB, GL_UNSIGNED_BYTE, image);
    delete []image;
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_NEAREST);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR_MIPMAP_NEAREST);

    return tex_num[num];
}


/*
 *
 */
void tveTexture::showLogo()
{
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluOrtho2D(0, 1, 0, 1);
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    glEnable(GL_TEXTURE_2D);
    glColor3f(1.0f, 1.0f, 1.0f);
    GLuint tex = load("logo.tex");
    glBegin(GL_TRIANGLE_STRIP);
        glTexCoord2f(0.0f, 0.0f);  glVertex2f( 0.0f, 0.0f);
        glTexCoord2f(0.0f, 1.0f);  glVertex2f( 0.0f, 1.0f);
        glTexCoord2f(1.0f, 0.0f);  glVertex2f( 1.0f, 0.0f);
        glTexCoord2f(1.0f, 1.0f);  glVertex2f( 1.0f, 1.0f);
    glEnd();
    SDL_GL_SwapBuffers();
    glDeleteTextures(1, &tex);
}


GLuint tveTexture::generatySky()
{
    int     i, j;
    GLuint  tex_num;
    tveRGBA *texture = new tveRGBA[64*64];

    // Generate sky texture
    for (i=0; i<64*64; i++)
    {
        texture[i][RED]   = 128+SRANDOM(32);
        texture[i][GREEN] = 160;
        texture[i][BLUE]  = 216+SRANDOM(16);
        texture[i][ALPHA] = 0xa0;
    }
    for (i=1; i<63; i++)
    {
        for (j=1; j<63; j++)
        {
            if (!RANDOM(32))
            {
                for (int k=-1; k<2; k++)
                {
                    for (int l=-1; l<2; l++)
                    {
                        texture[(i+k)*64+j+l][RED]   = MIN(texture[(i+k)*64+j+l][RED]+48,   0xff);
                        texture[(i+k)*64+j+l][GREEN] = MIN(texture[(i+k)*64+j+l][GREEN]+32, 0xff);
                        texture[(i+k)*64+j+l][BLUE]  = MIN(texture[(i+k)*64+j+l][BLUE]+20,  0xff);
                    }
                }
                texture[i*64+j][RED]   = MIN(texture[i*64+j][RED]+48,   0xff);
                texture[i*64+j][GREEN] = MIN(texture[i*64+j][GREEN]+32, 0xff);
                texture[i*64+j][BLUE]  = MIN(texture[i*64+j][BLUE]+20,  0xff);
            }
        }
    }
    for (i=1; i<63; i++)
    {
        for (j=1; j<63; j++)
        {
            int idx = (i+SRANDOM(3))*64+j+SRANDOM(3);
            texture[i*64+j][RED]   = MIN(texture[idx][RED], 0xff);
            texture[i*64+j][GREEN] = MIN(texture[idx][GREEN], 0xff);
            texture[i*64+j][BLUE]  = MIN(texture[idx][BLUE], 0xff);
        }
    }
    glGenTextures(1, &tex_num);
    glBindTexture(GL_TEXTURE_2D, tex_num);
    glPixelStorei(GL_UNPACK_ALIGNMENT, 4);
    // define the texture image and parameters
    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, 64, 64, 0, GL_RGBA, GL_UNSIGNED_BYTE, texture);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    delete []texture;

    return tex_num;
}
