/*
 * Copyright (c) 2001, Thierry Van Elsuw 
 * All rights reserved. 
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions 
 * are met:
 * 
 * Redistributions of source code must retain the above copyright 
 * notice, this list of conditions and the following disclaimer.
 * 
 * Redistributions in binary form must reproduce the above copyright 
 * notice, this list of conditions and the following disclaimer in the 
 * documentation and/or other materials provided with the distribution.
 * 
 * Neither the name of DS Improve nor the names of its contributors may 
 * be used to endorse or promote products derived from this software 
 * without specific prior written permission.
 * 
 * Module name   : tveTexture.h
 * Creation date : 16/02/2001
 * e-mail        : Thierry Van Elsuwe <tvelsuwe@swing.be>
 */

#ifndef _TVETEXTURE_H_
#define _TVETEXTURE_H_

#include "voyager.h"

typedef Uint8 tveRGBA[4];
typedef Uint8 tveRGB[3];

#ifndef IRIX
#pragma pack(1)
#endif
typedef struct {  // sizeof() = 18
  char   id_field_length;
  char   color_map_type;
  char   image_type;
  //Sint16 c_map_origin;
  //Sint16 c_map_length;
  char   dummy[4];
  char   c_map_size;
  Sint16 x_origin;
  Sint16 y_origin;
  Sint16 width;
  Sint16 height;
  char   image_pix_size;
  char   image_descriptor;
} tveTGAHeader;

typedef struct
{
    Uint8  manufacturer;   
    Uint8  version;        
    Uint8  encoding;    
    Uint8  bitsPerPixel;
    Sint16 xmin;
    Sint16 ymin;
    Sint16 xmax;
    Sint16 ymax;
    Sint16 hDpi;
    Sint16 vDpi;
    Uint8  colorMap[48];
    Uint8  reserved;    
    Uint8  numPlanes;
    /*
    Sint16 bytesPerLine;
    Sint16 paletteInfo; 
    Sint16 hScreenSize;  
    Sint16 vScreenSize;
    */
    Uint8 dummy[8];
    Uint8  filler[54];  
} tvePCXHeader;
#ifndef IRIX
#pragma pack()
#endif

#define RED   0
#define GREEN 1
#define BLUE  2
#define ALPHA 3

#define MAX_TEXTURE 1024

class tveTexture
{
private:
    static char    *tex_name[MAX_TEXTURE];
    static GLuint  tex_num[MAX_TEXTURE];
public:
    static void init();
    static void saveTGA(char *filename, int width, int height, tveRGB *buffer);
    static GLuint load(char *filename);
    static GLuint loada(char *filename, tveRGBA color);
    static GLuint loadPCX(char *filename);
    static GLuint generatySky();
    static void showLogo();
};

#endif
