/*
 * Copyright (c) 2001, Thierry Van Elsuw 
 * All rights reserved. 
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions 
 * are met:
 * 
 * Redistributions of source code must retain the above copyright 
 * notice, this list of conditions and the following disclaimer.
 * 
 * Redistributions in binary form must reproduce the above copyright 
 * notice, this list of conditions and the following disclaimer in the 
 * documentation and/or other materials provided with the distribution.
 * 
 * Neither the name of DS Improve nor the names of its contributors may 
 * be used to endorse or promote products derived from this software 
 * without specific prior written permission.
 * 
 * Module name   : voyager.cpp
 * Creation date : 14/02/2001
 * e-mail        : Thierry Van Elsuwe <tvelsuwe@swing.be>
 */

#include "voyager.h"
#include "tveTexture.h"
#include "tveClock.h"
#include "tveConsole.h"
#include "tveLandscape.h"
#include "tveDemo.h"
#include "tveModel.h"
#include <time.h>

static GLfloat cosinus[361];
static GLfloat sinus[361];

enum tveState
{
    ST_UNUSED   = 0x0010,
    ST_LOCK     = 0x0020,
    ST_YPOS     = 0x0040,
    ST_HELP     = 0x0080,
    ST_CULL     = 0x0100,
    ST_LINE     = 0x0200,
    ST_OVERLAY  = 0x0400,
};

static int  state     = /*ST_YPOS|ST_LOCK|*/ST_CULL|ST_OVERLAY;
tveClock    *my_clock = NULL;
tveConsole  *console  = NULL;
tveDemo     *demo     = NULL;
GLuint      tex_logo;

static tveLandscape *landscape = NULL;
static tveModel     *model = NULL;

#ifdef __WIN32__
PFNGLLOCKARRAYSEXTPROC          glLockArraysEXT;
PFNGLUNLOCKARRAYSEXTPROC        glUnlockArraysEXT;
PFNGLACTIVETEXTUREARBPROC       glActiveTextureARB;
PFNGLCLIENTACTIVETEXTUREARBPROC glClientActiveTextureARB;
PFNGLMULTITEXCOORD2FARBPROC     glMultiTexCoord2fARB;
#endif


/*
 * showHelp                                                                 
 */
void showHelp()
{
    int posx = (int)tveVoyager::cv_screenwidth->value.integer/4;
    int posy = (int)tveVoyager::cv_screenheight->value.integer/4;
    console->printf(posx, posy+275, 0xffFFffFF, 0.5f, "`: Show console");
    console->printf(posx, posy+250, 0xffFFffFF, 0.5f, "F1: Help On/Off");
    console->printf(posx, posy+225, 0xffFFffFF, 0.5f, "F2: Cycle rendering method");
    console->printf(posx, posy+200, 0xffFFffFF, 0.5f, "F3: Lock / Unlock");
    console->printf(posx, posy+175, 0xffFFffFF, 0.5f, "F4: Follow landscape");
    console->printf(posx, posy+150, 0xffFFffFF, 0.5f, "F5: Face culling On/Off");
    console->printf(posx, posy+125, 0xffFFffFF, 0.5f, "F6: Line drawing On/Off");
    console->printf(posx, posy+100, 0xffFFffFF, 0.5f, "Up/Down: forward/backward");
    console->printf(posx, posy+ 75, 0xffFFffFF, 0.5f, "Left/Right: turn left/right");
    console->printf(posx, posy+ 50, 0xffFFffFF, 0.5f, "Home/End: up/down");
    console->printf(posx, posy+ 25, 0xffFFffFF, 0.5f, "PgDw/PgUp: look up/down");
    console->printf(posx, posy+  0, 0xffFFffFF, 0.5f, "Mouse move: look around");
#ifdef DEMO_DS
    switch (state&0x0f)
    {
        case 0:
        case 1: console->printf(10, 10, 0xffFFffFF, 0.5f, "Method: %s", (state&ST_LOCK) ? "QuadTree TriStrip CVA" : "QuadTree TriStrip vertex array"); break;
    }
    console->printf((int)tveVoyager::cv_screenwidth->value.integer-170, 10, 0xffFFffFF, 0.5f, "Face culling %s", (state&ST_CULL) ? "On" : "Off");
#endif
}


/*
 * showLogo
 */
void showLogo()
{
    static float angle = 0.0f;

    angle = angle+tveVoyager::ratio*0.2f;
    if (angle > 360.0f)
    {
        angle -= 360.0f;
    }
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glOrtho(0, tveVoyager::cv_screenwidth->value.integer, 0, tveVoyager::cv_screenheight->value.integer, -200, 200);
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    glBindTexture(GL_TEXTURE_2D, tex_logo);
    glBlendFunc(GL_ONE, GL_ZERO);
    GLfloat size = tveVoyager::cv_screenwidth->value.integer/20.0f;
    glTranslatef(tveVoyager::cv_screenwidth->value.integer-size*1.5f, tveVoyager::cv_screenheight->value.integer-size*1.5f, 0);
    if ( !strcasecmp("yes", tveVoyager::cv_logoanim->value.text) )
    {
        glRotatef(angle, 0.0f, 1.0f, 0.0f);
    }
    glColor4f(1.0f, 1.0f, 1.0f, 1.0f);
    glBegin(GL_TRIANGLE_STRIP);
      glTexCoord2f(0.0f, 1.0f); glVertex2f(-size, -size);
      glTexCoord2f(0.0f, 0.0f); glVertex2f(-size,  size);
      glTexCoord2f(1.0f, 1.0f); glVertex2f( size, -size);
      glTexCoord2f(1.0f, 0.0f); glVertex2f( size,  size);
    glEnd();
}


/*
 * showOverlay
 */
void showOverlay()
{
    console->printf((int)tveVoyager::cv_screenwidth->value.integer-90, (int)tveVoyager::cv_screenheight->value.integer-20,
                    0xffFFffFF, 0.5f, "FPS: %d", my_clock->getCurrentFPS());
#ifndef DEMO_DS
    console->printf(10, (int)tveVoyager::cv_screenheight->value.integer-20, 0xffFFffFF, 0.5f, "x: %4.2f  z: %4.2f  y: %4.2f",
                    tveVoyager::cv_posx->value.floating, Z(tveVoyager::cv_posz->value.floating), tveVoyager::cv_posy->value.floating);
    switch (state&0x0f)
    {
        case 0:
        case 1: console->printf(10, 10, 0xffFFffFF, 0.5f, "Method: %s", (state&ST_LOCK) ? "QuadTree TriStrip CVA" : "QuadTree TriStrip vertex array"); break;
    }
#else
    console->printf(10, (int)tveVoyager::cv_screenheight->value.integer-20, 0xffFFffFF, 0.5f, "x: %4.0f  z: %4.0f  y: %3.0f",
                    tveVoyager::cv_posx->value.floating*(10.0f/SCALE_X), Z(tveVoyager::cv_posz->value.floating)*(10.0f/SCALE_Z), tveVoyager::cv_posy->value.floating/SCALE_Y);
#endif
    if (state&ST_HELP)
    {
        showHelp();
    }
#ifndef DEMO_DS
    console->printf((int)tveVoyager::cv_screenwidth->value.integer-170, 10, 0xffFFffFF, 0.5f, "Face culling %s", (state&ST_CULL) ? "On" : "Off");
#endif

    console->start();
    console->printBuffer();
    console->display();
    showLogo(); // must be called just before console->stop() as it destroys matrix
    console->stop();
}


/*
 *
 */
static void drawScene()
{
    glClear(GL_DEPTH_BUFFER_BIT|GL_COLOR_BUFFER_BIT);
    glEnable(GL_FOG);
    // camera movement
    glLoadIdentity(); // be sure  we use GL_MODELVIEW (formResize)
    glRotatef(tveVoyager::cv_anglex->value.floating, 1.0f, 0.0f, 0.0f); // rotation haut-bas
    glRotatef(tveVoyager::cv_angley->value.floating, 0.0f, 1.0f, 0.0f); // rotation gauche-droite

    // adjust the sky and draw it
    glDisable(GL_DEPTH_TEST);
    glDepthMask(GL_FALSE);
    glPushMatrix();
    glTranslatef(landscape->height_map->width/2-tveVoyager::cv_posx->value.floating/2.0f,
                 -tveVoyager::cv_posy->value.floating/1000.0f,
                 Z(-landscape->height_map->height/2-tveVoyager::cv_posz->value.floating/2.0f));
    landscape->drawSky();
    glPopMatrix();
    glDepthMask(GL_TRUE);
    glEnable(GL_DEPTH_TEST);

    glTranslatef(-tveVoyager::cv_posx->value.floating, -tveVoyager::cv_posy->value.floating, -Z(tveVoyager::cv_posz->value.floating));
    // end of camera movement
    //FIXME: remove that line
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    if (state&ST_CULL)
    {
        glEnable(GL_CULL_FACE);
    }
    else
    {
        glDisable(GL_CULL_FACE);
    }
    switch (state&0x0f)
    {
        case 0:
        case 1: landscape->drawQuadTreeVertexArray((state&ST_LOCK)!=0); break;
    }
    if (state&ST_CULL)
    {
        glDisable(GL_CULL_FACE);
    }
    landscape->drawWater();
    landscape->drawTrees();
    if (model)
    {
        model->render();
    }
    glDisable(GL_FOG);
    if (state&ST_OVERLAY)
    {
        showOverlay();
    }
    else
    {
        console->flushBuffer();
    }
    SDL_GL_SwapBuffers();
    my_clock->incFrame();
}


/*
 *
 */
static bool formCreate()
{
    GLfloat fogcolor[] = {0.6f, 0.7f, 0.8f, 1.0f};

#ifdef __WIN32__
    glActiveTextureARB       = (PFNGLACTIVETEXTUREARBPROC)      wglGetProcAddress("glActiveTextureARB");
    glClientActiveTextureARB = (PFNGLCLIENTACTIVETEXTUREARBPROC)wglGetProcAddress("glClientActiveTextureARB");
    glMultiTexCoord2fARB     = (PFNGLMULTITEXCOORD2FARBPROC)    wglGetProcAddress("glMultiTexCoord2fARB");
    if (!glActiveTextureARB || !glMultiTexCoord2fARB)
    {
        printf("Error: multitexture not supported.\n");
        return false;
    }
    glLockArraysEXT   = (PFNGLLOCKARRAYSEXTPROC)   wglGetProcAddress("glLockArraysEXT");
    glUnlockArraysEXT = (PFNGLUNLOCKARRAYSEXTPROC) wglGetProcAddress("glUnlockArraysEXT");
    if (!glLockArraysEXT || !glUnlockArraysEXT)
    {
        printf("Warning: compiled vertex array not supported.\n");
        state &= ~ST_LOCK; // remove lock if used
    }
#endif
    srand(time(NULL));
    // create sine and cosine table for movement computing
    for (int i=0; i<=360; i++)
    {
        cosinus[i] = (float)cos((NPI*i)/180.0f);
        sinus[i] = (float)sin((NPI*(float)i)/180.0f);
    }

    // default OpenGL environnement
    glShadeModel(GL_SMOOTH);
    glFogi(GL_FOG_MODE,GL_EXP2);
    glFogfv(GL_FOG_COLOR, fogcolor);
    glClearColor(fogcolor[0], fogcolor[1], fogcolor[2], fogcolor[3]);

    glCullFace(GL_BACK);
    glActiveTextureARB(GL_TEXTURE1_ARB);
    glEnable(GL_TEXTURE_2D);
    glActiveTextureARB(GL_TEXTURE0_ARB);
    glEnable(GL_TEXTURE_2D);
    glEnable(GL_BLEND);
    glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
    glDepthFunc(GL_LEQUAL);
    glEnable(GL_ALPHA_TEST);
    glAlphaFunc(GL_GREATER, 0.0f);

    tex_logo = tveTexture::load("ds.tex");
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

    console->exec("autoexec.cfg");
    // create a default landscape if not already done by autoexec
    if (!landscape)
    {
        landscape = new tveLandscape(NULL);
    }
    // create a default model if not already done by autoexec
#ifndef DEMO_DS
    if (!model)
    {
        model = new tveModel("droideka.mdz");
    }
    /*if (model)
    {
        model->print();
    }*/
#endif
    return true;
}


/*
 *
 */
static void formResize(GLsizei width, GLsizei height)
{
    tveVoyager::cv_screenwidth->value.integer  = width;
    tveVoyager::cv_screenheight->value.integer = height;
    glViewport(0, 0, width, height);
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluPerspective(tveVoyager::cv_fov->value.floating,  // FOV
                   width/(float)height,         // aspect ratio
                   0.4f,//tveVoyager::cv_far->value/1000.0f, // distance to near clipping plane
                   tveVoyager::cv_far->value.floating); // distance to far clipping plane
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    glMatrixMode(GL_MODELVIEW);
    glFogf(GL_FOG_DENSITY, tveVoyager::cv_fog->value.floating/tveVoyager::cv_far->value.floating);
    console->resize(width, height, (2*height)/5);
}


/*
 *
 */
bool quitCmd(void *obj, int tag, char *args)
{
    SDL_Event event;

    event.type=SDL_QUIT;
    SDL_PushEvent(&event);
    return true;
}


/*
 *
 */
bool loadCmd(void *obj, int tag, char *args)
{
    char filename[256];

    if (tag==0)
    {
        sprintf(filename, "../data/heightmaps/%s", args);
        if ( !atoi(args) && ((!strlen(args)) || !tveVoyager::fileExists(filename)) )
        {
            console->printf("Error: file not found.");
            return true;
        }
        delete landscape;
        landscape = new tveLandscape(args);
    }
    else
    {
        sprintf(filename, "../data/models/%s", args);
        if ( !atoi(args) && ((!strlen(args)) || !tveVoyager::fileExists(filename)) )
        {
            console->printf("Error: file not found.");
            return true;
        }
        delete model;
        model = new tveModel(args);
    }
    return true;
}


/*
 *
 */
bool screenshotCmd(void *obj, int tag, char *args)
{
    tveRGB *buffer;
    char   tganame[] = "../data/screenshots/memoire00.tga";
    long   height = (long)tveVoyager::cv_screenheight->value.integer;
    long   width  = (long)tveVoyager::cv_screenwidth->value.integer;

    buffer = new tveRGB[height*width];
    glReadPixels(0, 0, width, height, GL_RGB, GL_UNSIGNED_BYTE, buffer);
    for (int i=0; i<100; i++)
    {
        tganame[27] = i/10  + '0';
        tganame[28] = i%10  + '0';
        if (!tveVoyager::fileExists(tganame))
        {
            break;
        }
    }
    tveTexture::saveTGA(tganame, width, height, buffer);
    delete []buffer;
    return true;
}


/*
 *
 */
bool catCmd(void *obj, int tag, char *args)
{
    gzFile f;
    char    buffer[256];
    
    if ( !(f=tveVoyager::openFile(args, "rt")) )
    {
        console->printf("Error: file not found.");
        return true;
    }
    while (gzgets(f, buffer, sizeof(buffer)))
    {
        console->printf("%s", buffer);
    }
    gzclose(f);

    return true;
}


/*
 *
 */
bool handleDemo(void *obj, int tag, char *args)
{
    switch (tag&0xffff)
    {
        case DEMO_RECORDING:
            demo->start(DEMO_RECORDING, args);
            break;
        case DEMO_STOP:
            demo->stop();
            break;
        case DEMO_PLAYING:
            demo->start(DEMO_PLAYING, args);
            break;
        case DEMO_LOOPING:
            demo->start(DEMO_LOOPING, args);
            break;
        case DEMO_PAUSE:
            demo->togglePause();
            break;
    }
    return true;
}


/*
 *
 */
void viewOnChange()
{
    formResize((GLsizei)tveVoyager::cv_screenwidth->value.integer, (GLsizei)tveVoyager::cv_screenheight->value.integer);
}


/*
 *
 */
static void registerConsole()
{
    console = new tveConsole(20);
    console->addCommand(new tveConsoleCommand("quit", 0, quitCmd));
    console->addCommand(new tveConsoleCommand("cat", 0, catCmd));
    console->addCommand(new tveConsoleCommand("screenshot", 0, screenshotCmd));
    console->addCommand(new tveConsoleCommand("load",    0, loadCmd));
    console->addCommand(new tveConsoleCommand("loadmd2", 1, loadCmd));
    console->addCommand(new tveConsoleCommand("record", DEMO_RECORDING|CT_DELAYED, handleDemo));
    console->addCommand(new tveConsoleCommand("play",   DEMO_PLAYING|CT_DELAYED,   handleDemo));
    console->addCommand(new tveConsoleCommand("loop",   DEMO_LOOPING|CT_DELAYED,   handleDemo));
    console->addCommand(new tveConsoleCommand("stop",   DEMO_STOP|CT_DELAYED,      handleDemo));
    console->addCommand(new tveConsoleCommand("pause",  DEMO_PAUSE|CT_DELAYED,     handleDemo));
    console->addVariable(tveVoyager::cv_alpha  = new tveConsoleVariable("alpha",  VT_TEXT,  "Off", NULL));
    console->addVariable(tveVoyager::cv_beta   = new tveConsoleVariable("beta",   VT_TEXT,  "Off", NULL));
    console->addVariable(tveVoyager::cv_gamma  = new tveConsoleVariable("gamma",  VT_TEXT,  "Off", NULL));
    console->addVariable(tveVoyager::cv_far    = new tveConsoleVariable("far",    VT_FLOAT, "200", viewOnChange));
    console->addVariable(tveVoyager::cv_fog    = new tveConsoleVariable("fog",    VT_FLOAT, "2.5", viewOnChange));
    console->addVariable(tveVoyager::cv_fov    = new tveConsoleVariable("fov",    VT_FLOAT, "60",  viewOnChange));
    console->addVariable(tveVoyager::cv_speed  = new tveConsoleVariable("speed",  VT_FLOAT, "20",  NULL));
    console->addVariable(tveVoyager::cv_water  = new tveConsoleVariable("water",  VT_FLOAT, "4.01", NULL));
    console->addVariable(tveVoyager::cv_detail = new tveConsoleVariable("detail", VT_FLOAT, "5000", NULL));
    console->addVariable(tveVoyager::cv_anglex = new tveConsoleVariable("anglex", VT_FLOAT, "0",   NULL));
    console->addVariable(tveVoyager::cv_angley = new tveConsoleVariable("angley", VT_FLOAT, "135", NULL));
    console->addVariable(tveVoyager::cv_posx   = new tveConsoleVariable("posx",   VT_FLOAT, "10",  NULL));
    console->addVariable(tveVoyager::cv_posy   = new tveConsoleVariable("posy",   VT_FLOAT, "10",  NULL));
    console->addVariable(tveVoyager::cv_posz   = new tveConsoleVariable("posz",   VT_FLOAT, "-10", NULL));
    console->addVariable(tveVoyager::cv_screenwidth  = new tveConsoleVariable("screenwidth",  VT_INTEGER, "800", viewOnChange));
    console->addVariable(tveVoyager::cv_screenheight = new tveConsoleVariable("screenheight", VT_INTEGER, "600", viewOnChange));
    console->addVariable(tveVoyager::cv_fullscreen   = new tveConsoleVariable("fullscreen",   VT_TEXT,    "No",  NULL));
    console->addVariable(tveVoyager::cv_logoanim     = new tveConsoleVariable("logoanim",     VT_TEXT,    "Yes", NULL));
    console->addVariable(tveVoyager::cv_consolekey   = new tveConsoleVariable("consolekey",   VT_INTEGER, "178", NULL));
    console->exec("config.cfg");
}


/*
 *
 */
static void handleKeyState()
{
    if (console->visible())
    {
        return; // do nothing if the console is active
    }

    float ratio = tveVoyager::ratio*0.001f;
    Uint8 *keystate = SDL_GetKeyState(NULL);

    if (keystate[SDLK_END])
    {
        tveVoyager::cv_posy->value.floating /= (1.0f+ratio*2.0f);
    }
    if (keystate[SDLK_HOME])
    {
        tveVoyager::cv_posy->value.floating = (tveVoyager::cv_posy->value.floating+0.01f)*(1.0f+ratio*2.0f);  // +0.01 so we are not stalled at 0
    }
    if (keystate[SDLK_PAGEUP])
    {
        tveVoyager::cv_anglex->value.floating = MIN(tveVoyager::cv_anglex->value.floating+ratio*50.0f, 90.0f); //TODO: put a rotation speed cvar
    }
    if (keystate[SDLK_PAGEDOWN])
    {
        tveVoyager::cv_anglex->value.floating = MAX(tveVoyager::cv_anglex->value.floating-ratio*50.0f, -90.0f); //TODO: put a rotation speed cvar
    }
    if (keystate[SDLK_UP])
    {
        tveVoyager::cv_posx->value.floating += sinus[(int)tveVoyager::cv_angley->value.floating]*ratio*tveVoyager::cv_speed->value.floating;
        tveVoyager::cv_posz->value.floating += cosinus[(int)tveVoyager::cv_angley->value.floating]*ratio*tveVoyager::cv_speed->value.floating;
    }
    if (keystate[SDLK_DOWN])
    {
        tveVoyager::cv_posx->value.floating -= sinus[(int)tveVoyager::cv_angley->value.floating]*ratio*tveVoyager::cv_speed->value.floating;
        tveVoyager::cv_posz->value.floating -= cosinus[(int)tveVoyager::cv_angley->value.floating]*ratio*tveVoyager::cv_speed->value.floating;
    }
    if (keystate[SDLK_LEFT])
    {
        if (SDL_GetModState() & (KMOD_LALT|KMOD_RALT))
        {
            int angle = ((int)tveVoyager::cv_angley->value.floating+270)%360;
            tveVoyager::cv_posx->value.floating += sinus[angle]*ratio*tveVoyager::cv_speed->value.floating;
            tveVoyager::cv_posz->value.floating += cosinus[angle]*ratio*tveVoyager::cv_speed->value.floating;
        }
        else
        {
            tveVoyager::cv_angley->value.floating -= 2.0f*ratio*50.0f; //TODO: put a rotation speed cvar
        }
    }
    if (keystate[SDLK_RIGHT])
    {
        if (SDL_GetModState() & (KMOD_LALT|KMOD_RALT))
        {
            int angle = ((int)tveVoyager::cv_angley->value.floating+90)%360;
            tveVoyager::cv_posx->value.floating += sinus[angle]*ratio*tveVoyager::cv_speed->value.floating;
            tveVoyager::cv_posz->value.floating += cosinus[angle]*ratio*tveVoyager::cv_speed->value.floating;
        }
        else
        {
            tveVoyager::cv_angley->value.floating += 2.0f*ratio*50.0f; //TODO: put a rotation speed cvar
        }
    }

    if (tveVoyager::cv_angley->value.floating < 0.0f)
    {
        tveVoyager::cv_angley->value.floating += 360.0f;
    }
    else if (tveVoyager::cv_angley->value.floating > 360.0f)
    {
        tveVoyager::cv_angley->value.floating -= 360.0f;
    }
}


/*
 *
 */
static void handleMouseMotion(SDL_MouseMotionEvent motion)
{
    tveVoyager::cv_anglex->value.floating -= motion.yrel*0.25f;
    if (tveVoyager::cv_anglex->value.floating > 90.0f)
    {
        tveVoyager::cv_anglex->value.floating = 90.0f;
    }
    else if (tveVoyager::cv_anglex->value.floating < -90.0f)
    {
        tveVoyager::cv_anglex->value.floating = -90.0f;
    }
    tveVoyager::cv_angley->value.floating += motion.xrel*0.25f;
    if (tveVoyager::cv_angley->value.floating < 0.0f)
    {
        tveVoyager::cv_angley->value.floating += 360.0f;
    }
    else if (tveVoyager::cv_angley->value.floating > 360.0f)
    {
        tveVoyager::cv_angley->value.floating -= 360.0f;
    }
}


/*
 *
 */
static void idleFunc()
{
    static bool was_under_water = false;
    static int  last_tick = SDL_GetTicks();
    int         current_tick = SDL_GetTicks();

    tveVoyager::ratio = (float)(current_tick-last_tick);
    last_tick = current_tick;

    handleKeyState();
    if (state&ST_YPOS)
    {
        //tveVoyager::cv_posy->value.floating = landscape->getYPos(tveVoyager::cv_posx->value.floating, Z(tveVoyager::cv_posz->value.floating))+0.5f;
        tveVoyager::cv_posy->value.floating += 0.5f * (landscape->getYPos(tveVoyager::cv_posx->value.floating, Z(tveVoyager::cv_posz->value.floating))+1.0f - tveVoyager::cv_posy->value.floating);
    }
    if (was_under_water == (tveVoyager::cv_posy->value.floating > tveVoyager::cv_water->value.floating) )
    {
        GLfloat fogcolor[] = {0.6f, 0.7f, 0.8f, 1.0f};

        was_under_water = !was_under_water;
        if (was_under_water)
        {
            fogcolor[0] = 0.0f;
            fogcolor[1] = 0.0f;
            fogcolor[2] = 0.1f;
            tveVoyager::cv_far->value.floating *= 0.1f;
            glFogi(GL_FOG_MODE, GL_EXP);
        }
        else
        {
            tveVoyager::cv_far->value.floating *= 10.0f;
            glFogi(GL_FOG_MODE, GL_EXP2);
        }
        glFogfv(GL_FOG_COLOR, fogcolor);
        glClearColor(fogcolor[0], fogcolor[1], fogcolor[2], fogcolor[3]);
        viewOnChange();
    }
    console->handleCommands();
    demo->execute();
    drawScene();
    //usleep(0);
}


/*
 *
 */
static bool initEngine(int argc, char *argv[])
{
    tveVoyager::init(argc, argv, NULL);
    registerConsole();
    if ( SDL_SetVideoMode((int)tveVoyager::cv_screenwidth->value.integer, (int)tveVoyager::cv_screenheight->value.integer,
                          0, SDL_OPENGL | (!strcasecmp("yes", tveVoyager::cv_fullscreen->value.text) ? SDL_FULLSCREEN : 0)) == NULL )
    {
        fprintf(stderr, "Unable to create OpenGL screen: %s\n", SDL_GetError());
        return false;
    }
    tveTexture::showLogo();
    printf("%s\n", glGetString(GL_RENDERER));
    printf("%s\n", glGetString(GL_VERSION));
    printf("%s\n", glGetString(GL_VENDOR));
    SDL_WM_SetCaption("Voyager (c) 2001, Thierry Van Elsuwe (tvelsuwe@swing.be)", NULL);
    if (!formCreate())
    {
        return false;
    }
    formResize((GLsizei)tveVoyager::cv_screenwidth->value.integer, (GLsizei)tveVoyager::cv_screenheight->value.integer);

    SDL_ShowCursor(0);
    SDL_WM_GrabInput(SDL_GRAB_ON);
    while (SDL_PollEvent(NULL)) // flush events queue
    {
        SDL_Event event;
        SDL_WaitEvent(&event);
    }

    my_clock = new tveClock();
    demo = new tveDemo();

    return true;
}


/*
 *
 */
static void mainLoop()
{
    for ( ; ; )
    {
        while (!SDL_PollEvent(NULL)) idleFunc();
        SDL_Event event;
        SDL_WaitEvent(&event);
        switch (event.type)
        {
            case SDL_MOUSEMOTION: handleMouseMotion(event.motion); break;

            case SDL_MOUSEBUTTONDOWN: break;

            case SDL_KEYDOWN:
                console->handleKeyDown(event.key.keysym);
                if (event.key.keysym.sym == SDLK_F2)
                {
                    state = ((state+1)&0x0001) | (state&0xfff0);
                }
                else if (event.key.keysym.sym == SDLK_F3)
                {
                    state ^= ST_LOCK;
                }
                else if (event.key.keysym.sym == SDLK_F4)
                {
                    state ^= ST_YPOS;
                }
                else if (event.key.keysym.sym == SDLK_F5)
                {
                    state ^= ST_CULL;
                }
                else if (event.key.keysym.sym == SDLK_F6)
                {
                    state ^= ST_LINE;
                    if (state&ST_LINE)
                        glPolygonMode(GL_FRONT, GL_LINE);
                    else
                        glPolygonMode(GL_FRONT, GL_FILL);
                }
                else if (event.key.keysym.sym == SDLK_F7)
                {
                    state ^= ST_OVERLAY;
                }
                else if ( (event.key.keysym.sym==SDLK_F10) || (event.key.keysym.sym==SDLK_ESCAPE && !console->visible()) )
                {
                    quitCmd(NULL, 0, NULL);
                }
                if (console->visible()) break; // do nothing if the console is active
                if ( (event.key.keysym.sym==SDLK_h) || (event.key.keysym.sym==SDLK_F1) )
                {
                    state ^= ST_HELP;
                }
                else if (event.key.keysym.sym == SDLK_s)
                {
                    screenshotCmd(NULL, 0, NULL);
                }
                break;

            case SDL_QUIT:
                printf("Nombre approximatif d'images par seconde: %u fps\n", my_clock->getTotalFPS());
                return;
        }
    }
}


/*
 *
 */
void quitEngine()
{
    console->saveConfig();
}


/*
 *
 */
int main(int argc, char *argv[])
{
    printf("\n--------------------------------------------------------------------\n");
    printf(" Terrain Voyager Engine - Thierry Van Elsuwe - tvelsuwe@swing.be\n");
    printf("--------------------------------------------------------------------\n\n");
    // Initialize SDL for video output and timer
    if ( SDL_Init(SDL_INIT_VIDEO|SDL_INIT_TIMER/*|SDL_INIT_NOPARACHUTE*/) < 0 )
    {
        fprintf(stderr, "Unable to initialize SDL: %s\n", SDL_GetError());
        return 1;
    }
    // Initialize voyager engine
    if ( !initEngine(argc, argv) )
    {
        fprintf(stderr, "Unable to initialize Voyager engine\n");
        return 1;
    }
    mainLoop();
    quitEngine();
    SDL_Quit();
    return 0;
}
